from beartype import beartype
import numpy as np
import random
import torch
import time
import os

from ..onnx2pytorch.convert.model import ConvertModel
from ..verifier.objective import DnfObjectives

class RandomAttacker:

    # @beartype
    def __init__(self: 'RandomAttacker', net: ConvertModel, objective: DnfObjectives, input_shape: tuple, device: str) -> None:
        self.net = net
        self.objective = objective
        self.input_shape = input_shape
        self.device = device

        if np.prod(self.input_shape) >= 200:
            return None

        self.n_runs = 10
        self.n_samples = 50
        self.n_pos_samples = int(self.n_samples * 0.1)

        self.output_shape = self.net(torch.zeros(self.input_shape, device=device)).shape
        self.target, self.direction = self.get_target_and_direction()

        # print(self.target)
        # print(self.direction)


    # @beartype
    def manual_seed(self: 'RandomAttacker', seed: int) -> None:
        self.seed = seed
        random.seed(self.seed)
        torch.manual_seed(self.seed)


    # @beartype
    def get_target_and_direction(self: 'RandomAttacker') -> tuple[int, str]:
        target_dict = dict.fromkeys(range(np.prod(self.output_shape)), 0)
        obj_dict = dict.fromkeys(range(np.prod(self.output_shape)), 0)
        for arr in self.objective.cs:
            for k in range(len(arr)):
                for kk in range(len(arr[k])):
                    if (arr[k][kk] != 0):
                        target_dict[kk] += 1
                    if (arr[k][kk] < 0):
                        obj_dict[kk] += 1

        target = sorted(target_dict.items(), key=lambda item: item[1])[-1][0]
        obj_type = sorted(obj_dict.items(), key=lambda item: item[1])[-1][0]
        if target == obj_type:
            direction = 'maximize'
        else:
            direction = 'minimize'

        return target, direction


    # @beartype
    def _attack(self: 'RandomAttacker', input_lowers: torch.Tensor, input_uppers: torch.Tensor) -> torch.Tensor | None:
        adv = self._sampling(
            input_lowers=input_lowers,
            input_uppers=input_uppers,
            target=self.target,
            direction=self.direction,
        )
        return adv


    # @beartype
    def run(self: 'RandomAttacker', timeout: float = 1.0) -> tuple[bool, torch.Tensor | None]:
        if np.prod(self.input_shape) >= 200:
            return False, None

        indexes = torch.randperm(self.objective.lower_bounds.shape[0])
        shuffled_lower_bounds = self.objective.lower_bounds[indexes]
        shuffled_upper_bounds = self.objective.upper_bounds[indexes]
        if not torch.equal(self.objective.lower_bounds, shuffled_lower_bounds) or \
            not torch.equal(self.objective.upper_bounds, shuffled_upper_bounds):
                return False, None

        input_lowers = self.objective.lower_bounds[0].clone().view(self.input_shape).to(self.device)
        input_uppers = self.objective.upper_bounds[0].clone().view(self.input_shape).to(self.device)

        start = time.time()
        while True:
            adv = self._attack(input_lowers=input_lowers, input_uppers=input_uppers)
            if adv is not None:
                return True, adv

            if time.time() - start > timeout:
                return False, None


    # @beartype
    def _sampling(self: 'RandomAttacker', input_lowers: torch.Tensor, input_uppers: torch.Tensor, target: int, direction: str) -> torch.Tensor | None:
        old_pos_samples = []

        for it in range(self.n_runs):
            stat, samples = self._make_samples(
                input_lowers=input_lowers,
                input_uppers=input_uppers,
            )
            if stat:
                return samples[0][0]

            pos_samples, neg_samples = self._split_samples(
                samples=samples,
                target=target,
                direction=direction,
            )

            if len(old_pos_samples) > 0:
                pos_samples, neg_samples_2 = self._split_samples(
                    samples=pos_samples + old_pos_samples,
                    target=target,
                    direction=direction,
                )
                neg_samples = neg_samples_2 + neg_samples

            old_pos_samples = pos_samples

            if torch.all((input_uppers - input_lowers) <= 1e-6):
                return None

            input_lowers, input_uppers = self._learning(
                input_lowers=input_lowers,
                input_uppers=input_uppers,
                pos_samples=pos_samples,
                neg_samples=neg_samples,
            )

        return None


    # @beartype
    def _learning(self: 'RandomAttacker', input_lowers: torch.Tensor, input_uppers: torch.Tensor,
                  pos_samples: list[tuple[torch.Tensor, torch.Tensor]], neg_samples: list[tuple[torch.Tensor, torch.Tensor]]) -> tuple[torch.Tensor, torch.Tensor]:
        new_input_lowers = input_lowers.clone().flatten()
        new_input_uppers = input_uppers.clone().flatten()
        random.shuffle(neg_samples)

        pos_sample = random.choice(pos_samples)
        dim = random.randint(0, new_input_uppers.numel() - 1)
        for neg_sample in neg_samples:
            pos_val = pos_sample[0][dim]
            neg_val = neg_sample[0][dim]
            if pos_val > neg_val:
                temp = random.uniform(neg_val, pos_val)
                if new_input_lowers[dim] <= temp <= new_input_uppers[dim]:
                    new_input_lowers[dim] = temp
            else:
                temp = random.uniform(pos_val, neg_val)
                if new_input_lowers[dim] <= temp <= new_input_uppers[dim]:
                    new_input_uppers[dim] = temp

        return new_input_lowers.view(self.input_shape), new_input_uppers.view(self.input_shape)


    # @beartype
    def _split_samples(self: 'RandomAttacker', samples: list[tuple[torch.Tensor, torch.Tensor]], target: int, direction: str) -> tuple:
        if direction == 'minimize':
            sorted_samples = sorted(samples, key=lambda tup: tup[1][target])
        else:
            sorted_samples = sorted(samples, key=lambda tup: tup[1][target], reverse=True)

        pos_samples = sorted_samples[:self.n_pos_samples]
        neg_samples = sorted_samples[self.n_pos_samples:]

        return pos_samples, neg_samples


    # @beartype
    def _make_samples(self: 'RandomAttacker', input_lowers: torch.Tensor, input_uppers: torch.Tensor) -> tuple[bool, list[tuple[torch.Tensor, torch.Tensor]]]:
        s_in = (input_uppers - input_lowers) * torch.rand(self.n_samples, *self.input_shape[1:], device=self.device) + input_lowers
        if os.environ.get('NEURALSAT_ASSERT'):
            assert torch.all(input_lowers <= s_in) and torch.all(s_in <= input_uppers)
        s_out = self.net(s_in)
        samples = []
        for cs, rhs in zip(self.objective.cs.to(self.device), self.objective.rhs.to(self.device)):
            vec = cs @ s_out.transpose(0, 1)
            for i in range(self.n_samples):
                if torch.all(vec[:, i] <= rhs):
                    return True, [(s_in[i].view(self.input_shape), s_out[i].flatten())]
                sample = s_in[i].flatten(), s_out[i].flatten()
                samples.append(sample)
        return False, samples


    def __repr__(self):
        return f'RandomAttack(seed={self.seed}, device={self.device})'



from beartype import beartype
import numpy as np
import torch

from ..util.network.read_onnx import custom_quirks


class Objective:

    "Single objective in CNF"

    # @beartype
    def __init__(self: 'Objective', prop: tuple[list, tuple[np.ndarray, np.ndarray]]) -> None:
        input_bounds, mat = prop
        self.dtype = torch.get_default_dtype()

        bounds = torch.tensor(input_bounds, dtype=self.dtype)
        self.lower_bound = bounds[:, 0]
        self.upper_bound = bounds[:, 1]
        assert torch.all(self.lower_bound <= self.upper_bound)

        bounds_f64 = torch.tensor(input_bounds, dtype=torch.float64)
        self.lower_bound_f64 = bounds_f64[:, 0]
        self.upper_bound_f64 = bounds_f64[:, 1]
        assert torch.all(self.lower_bound_f64 <= self.upper_bound_f64)

        self._extract(mat)


    # @beartype
    def _extract(self: 'Objective', mat: tuple[np.ndarray, np.ndarray]) -> None:
        assert len(mat) == 2, print(len(mat))
        prop_mat, prop_rhs = mat

        # f32
        self.cs = torch.tensor(prop_mat, dtype=self.dtype)
        self.rhs = torch.tensor(prop_rhs, dtype=self.dtype)

        # f64
        self.cs_f64 = torch.tensor(prop_mat, dtype=torch.float64)
        self.rhs_f64 = torch.tensor(prop_rhs, dtype=torch.float64)

        if custom_quirks.get('Softmax', {}).get('skip_last_layer', False):
            assert (self.rhs == 0).all()


    # @beartype
    def get_info(self) -> tuple[torch.Tensor, torch.Tensor]:
        return self.cs, self.rhs


    # @beartype
    def get_info_f64(self) -> tuple[torch.Tensor, torch.Tensor]:
        return self.cs_f64, self.rhs_f64


class DnfObjectives:

    "List of objectives"

    # @beartype
    def __init__(self, objectives: list[Objective], input_shape: tuple, is_nhwc: bool) -> None:
        self.objectives = objectives
        self.is_nhwc = is_nhwc
        self.input_shape = input_shape

        self._extract()

        self.num_used = 0


    # @beartype
    def __len__(self: 'DnfObjectives') -> int:
        return len(self.lower_bounds[self.num_used:])


    # @beartype
    def pop(self: 'DnfObjectives', batch: int):
        if isinstance(self.cs, torch.Tensor):
            batch = min(batch, len(self))
        else:
            batch = 1
        # print('\t- popping:', batch)
        class TMP:
            pass

        assert len(self.lower_bounds) == len(self.upper_bounds) == len(self.cs) == len(self.rhs)
        assert len(self.lower_bounds_f64) == len(self.upper_bounds_f64) == len(self.cs_f64) == len(self.rhs_f64)
        assert len(self.cs) == len(self.cs_f64)

        # f32
        lower_bounds = self.lower_bounds[self.num_used : self.num_used + batch]
        upper_bounds = self.upper_bounds[self.num_used : self.num_used + batch]

        # f64
        lower_bounds_f64 = self.lower_bounds_f64[self.num_used : self.num_used + batch]
        upper_bounds_f64 = self.upper_bounds_f64[self.num_used : self.num_used + batch]

        objective = TMP()

        # indices for distinguishing restart
        objective.ids = torch.arange(0, batch) + self.num_used + 3

        # input bounds
        objective.lower_bounds = lower_bounds
        objective.upper_bounds = upper_bounds

        objective.lower_bounds_f64 = lower_bounds_f64
        objective.upper_bounds_f64 = upper_bounds_f64

        # specs
        objective.cs = self.cs[self.num_used : self.num_used + batch]
        objective.cs_f64 = self.cs_f64[self.num_used : self.num_used + batch]
        if not isinstance(objective.cs, torch.Tensor):
            objective.cs = torch.cat(objective.cs)[None]
            objective.cs_f64 = torch.cat(objective.cs_f64)[None]

        objective.rhs = self.rhs[self.num_used : self.num_used + batch]
        objective.rhs_f64 = self.rhs_f64[self.num_used : self.num_used + batch]
        if not isinstance(objective.rhs, torch.Tensor):
            objective.rhs = torch.cat(objective.rhs)[None]
            objective.rhs_f64 = torch.cat(objective.rhs_f64)[None]

        self.num_used += batch

        assert objective.cs.dtype == objective.rhs.dtype == objective.lower_bounds.dtype == objective.upper_bounds.dtype
        assert objective.cs_f64.dtype == objective.rhs_f64.dtype == objective.lower_bounds_f64.dtype == objective.upper_bounds_f64.dtype

        return objective


    # @beartype
    def _extract(self: 'DnfObjectives') -> None:
        self.cs, self.rhs = [], []
        self.lower_bounds, self.upper_bounds = [], []

        self.cs_f64, self.rhs_f64 = [], []
        self.lower_bounds_f64, self.upper_bounds_f64 = [], []

        for objective in self.objectives:
            # f32
            self.lower_bounds.append(objective.lower_bound)
            self.upper_bounds.append(objective.upper_bound)

            c_, rhs_ = objective.get_info()
            self.cs.append(c_)
            self.rhs.append(rhs_)

            # f64
            self.lower_bounds_f64.append(objective.lower_bound_f64)
            self.upper_bounds_f64.append(objective.upper_bound_f64)

            c_f64, rhs_f64 = objective.get_info_f64()
            self.cs_f64.append(c_f64)
            self.rhs_f64.append(rhs_f64)

        # input bounds
        self.lower_bounds = torch.stack(self.lower_bounds)
        self.upper_bounds = torch.stack(self.upper_bounds)

        self.lower_bounds_f64 = torch.stack(self.lower_bounds_f64)
        self.upper_bounds_f64 = torch.stack(self.upper_bounds_f64)

        if self.is_nhwc:
            n_, c_, h_, w_ = self.input_shape
            orig_input_shape = (-1, h_, w_, c_)
            # f32
            self.lower_bounds = self.lower_bounds.view(orig_input_shape).permute(0, 3, 1, 2).flatten(1)
            self.upper_bounds = self.upper_bounds.view(orig_input_shape).permute(0, 3, 1, 2).flatten(1)

            # f64
            self.lower_bounds_f64 = self.lower_bounds_f64.view(orig_input_shape).permute(0, 3, 1, 2).flatten(1)
            self.upper_bounds_f64 = self.upper_bounds_f64.view(orig_input_shape).permute(0, 3, 1, 2).flatten(1)

        assert torch.all(self.lower_bounds <= self.upper_bounds)
        assert torch.all(self.lower_bounds_f64 <= self.upper_bounds_f64)

        # properties
        if all([_.shape[0] == self.cs[0].shape[0] for _ in self.cs]):
            self.cs = torch.stack(self.cs)
            self.cs_f64 = torch.stack(self.cs_f64)
        if all([_.shape[0] == self.rhs[0].shape[0] for _ in self.rhs]):
            self.rhs = torch.stack(self.rhs)
            self.rhs_f64 = torch.stack(self.rhs_f64)


    # @beartype
    def add(self: 'DnfObjectives', objective) -> None:
        self.num_used -= len(objective.cs)



from __future__ import annotations
from collections import defaultdict
from beartype import beartype
import numpy as np
import typing
import random
import torch
import os

if typing.TYPE_CHECKING:
    from .. import abstractor

from ..heuristic.util import _compute_babsr_scores
from ..util.misc.result import AbstractResults
from ..setting import Settings

LARGE = 1e6
SMALL = 1.0 / LARGE

class DecisionHeuristic:

    # @beartype
    def __init__(self, input_split: bool, decision_topk: int, decision_method: str) -> None:
        self.input_split = input_split
        self.decision_topk = decision_topk
        self.decision_method = decision_method
        self.decision_reduceop = torch.max


    # @beartype
    @torch.no_grad()
    def __call__(self: 'DecisionHeuristic', abstractor: 'abstractor.abstractor.NetworkAbstractor', domain_params: AbstractResults) -> torch.Tensor | list[list]:
        if Settings.test:
            # hidden split
            return self.naive_hidden_branching(
                domain_params=domain_params,
                abstractor=abstractor,
                mode=random.choice(['scale', 'distance', 'polarity']),
            )


        if self.input_split:
            if (self.decision_method == 'smart') and (domain_params.lAs is not None):
                return self.smart_input_branching(
                    domain_params=domain_params,
                    abstractor=abstractor,
                )

            return self.naive_input_branching(
                domain_params=domain_params,
                abstractor=abstractor,
            )

        if None in abstractor.split_points:
            # FIXME: generalize it with relu
            # handle general activation
            return self.naive_hidden_branching(
                domain_params=domain_params,
                abstractor=abstractor,
                mode=random.choice(['width', 'distance'][1:]),
            )

        # hidden split
        if self.decision_method == 'greedy':
            return self.brute_force_hidden_branching(
                abstractor=abstractor,
                domain_params=domain_params,
            )

        if self.decision_method != 'smart':
            if random.uniform(0, 1) > 0.7:
                return self.naive_hidden_branching(
                    domain_params=domain_params,
                    abstractor=abstractor,
                    mode=random.choice(['scale', 'distance', 'polarity'])
                )

        return self.smart_hidden_branching(
            abstractor=abstractor,
            domain_params=domain_params,
        )


    # @beartype
    def get_topk_scores(self: 'DecisionHeuristic', abstractor: 'abstractor.abstractor.NetworkAbstractor', domain_params: AbstractResults,
                        topk_scores: torch.return_types.topk, topk_backup_scores: torch.return_types.topk, score_length: np.ndarray,
                        topk: int) -> tuple[torch.Tensor, list]:

        topk_decisions = []
        batch = len(domain_params.input_lowers)
        topk_output_lbs = torch.empty(
            size=(topk, batch * 2),
            device=domain_params.input_lowers.device,
            requires_grad=False,
        )

        # hidden
        double_lower_bounds = {k: torch.cat([v, v]) for k, v in domain_params.lower_bounds.items()}
        double_upper_bounds = {k: torch.cat([v, v]) for k, v in domain_params.upper_bounds.items()}

        # slope
        double_slopes = defaultdict(dict)
        for k, v in domain_params.slopes.items():
            double_slopes[k] = {kk: torch.cat([vv, vv], dim=2) for (kk, vv) in v.items()}

        # spec
        double_cs = torch.cat([domain_params.cs, domain_params.cs])
        double_rhs = torch.cat([domain_params.rhs, domain_params.rhs])

        # input
        double_input_lowers = torch.cat([domain_params.input_lowers, domain_params.input_lowers])
        double_input_uppers = torch.cat([domain_params.input_uppers, domain_params.input_uppers])

        if os.environ.get('NEURALSAT_ASSERT'):
            assert torch.all(double_input_lowers <= double_input_uppers)

        topk_scores_indices = topk_scores.indices.cpu()
        topk_backup_scores_indices = topk_backup_scores.indices.cpu()

        for k in range(topk):
            # top-k candidates from scores
            decision_max = [] # higher is better
            for idx in topk_scores_indices[:, k]:
                idx = idx.item()
                layer_idx = np.searchsorted(score_length, idx, side='right') - 1
                layer_name = abstractor.net.split_nodes[layer_idx].name
                layer_split_point = abstractor.net.split_activations[layer_name][0][0].get_split_point()
                neuron_idx = idx - score_length[layer_idx]
                if layer_split_point is not None: # relu
                    decision_max.append([layer_name, neuron_idx, layer_split_point])
                else: # general activation
                    raise NotImplementedError

            # print(f'{decision_max=} {len(decision_max)=}')

            # top-k candidates from backup scores.
            decision_min = [] # lower is better
            for idx in topk_backup_scores_indices[:, k]:
                idx = idx.item()
                layer_idx = np.searchsorted(score_length, idx, side='right') - 1
                layer_name = abstractor.net.split_nodes[layer_idx].name
                layer_split_point = abstractor.net.split_activations[layer_name][0][0].get_split_point()
                neuron_idx = idx - score_length[layer_idx]
                if layer_split_point is not None: # relu
                    decision_min.append([layer_name, neuron_idx, layer_split_point])
                else: # general activation
                    raise NotImplementedError

            # top-k candidates
            topk_decisions.append(decision_max + decision_min)

            k_domain_params = AbstractResults(**{
                'input_lowers': double_input_lowers,
                'input_uppers': double_input_uppers,
                'lower_bounds': double_lower_bounds,
                'upper_bounds': double_upper_bounds,
                'slopes': double_slopes if k == 0 else [],
                'cs': double_cs,
                'rhs': double_rhs,
            })

            abs_ret = abstractor._forward_hidden(
                domain_params=k_domain_params,
                decisions=topk_decisions[-1],
                simplify=True
            )
            # improvements over specification
            k_output_lbs = (abs_ret.output_lbs - torch.cat([double_rhs, double_rhs])).max(-1).values

            # invalid scores for stable neurons
            invalid_mask_scores = (topk_scores.values[:, k] <= SMALL).to(torch.get_default_dtype())
            invalid_mask_backup_scores = (topk_backup_scores.values[:, k] >= -SMALL).to(torch.get_default_dtype())
            invalid_mask = torch.cat([invalid_mask_scores, invalid_mask_backup_scores]).repeat(2) * LARGE
            topk_output_lbs[k] = self.decision_reduceop((k_output_lbs.view(-1) - invalid_mask).reshape(2, -1), dim=0).values

        return topk_output_lbs, topk_decisions


    # @beartype
    def get_topk_scores_greedy(self: 'DecisionHeuristic', abstractor: 'abstractor.abstractor.NetworkAbstractor', domain_params: AbstractResults,
                        topk_scores: torch.return_types.topk, score_length: np.ndarray, topk: int, reduce_op=torch.max) -> tuple[torch.Tensor, list]:
        topk_decisions = []
        batch = len(domain_params.input_lowers)
        topk_output_lbs = torch.empty(
            size=(topk, batch),
            device=domain_params.input_lowers.device,
            requires_grad=False,
        )

        # # hidden
        # double_lower_bounds = {k: torch.cat([v, v]) for k, v in domain_params.lower_bounds.items()}
        # double_upper_bounds = {k: torch.cat([v, v]) for k, v in domain_params.upper_bounds.items()}

        # # slope
        # double_slopes = defaultdict(dict)
        # for k, v in domain_params.slopes.items():
        #     double_slopes[k] = {kk: torch.cat([vv, vv], dim=2) for (kk, vv) in v.items()}

        # # spec
        # double_cs = torch.cat([domain_params.cs, domain_params.cs])
        double_rhs = torch.cat([domain_params.rhs, domain_params.rhs])

        # # input
        # double_input_lowers = torch.cat([domain_params.input_lowers, domain_params.input_lowers])
        # double_input_uppers = torch.cat([domain_params.input_uppers, domain_params.input_uppers])

        # if os.environ.get('NEURALSAT_ASSERT'):
        #     assert torch.all(double_input_lowers <= double_input_uppers)

        topk_scores_indices = topk_scores.indices.cpu()
        topk_scores_values = topk_scores.values.cpu()

        for k in range(topk):
            # top-k candidates from scores
            decision_candidates = [] # higher is better
            for idx, value in zip(topk_scores_indices[:, k], topk_scores_values[:, k]):
                if value == 0.0:
                    decision_candidates.append(topk_decisions[0][0])
                    print(idx)
                    continue
                idx = idx.item()
                layer_idx = np.searchsorted(score_length, idx, side='right') - 1
                layer_name = abstractor.net.split_nodes[layer_idx].name
                layer_split_point = abstractor.net.split_activations[layer_name][0][0].get_split_point()
                neuron_idx = idx - score_length[layer_idx]
                if layer_split_point is not None: # relu
                    decision_candidates.append([layer_name, neuron_idx, layer_split_point])
                else: # general activation
                    raise NotImplementedError

            # print(f'{decision_candidates=} {len(decision_candidates)=}')

            # top-k candidates
            topk_decisions.append(decision_candidates)

            k_domain_params = AbstractResults(**{
                'input_lowers': domain_params.input_lowers,
                'input_uppers': domain_params.input_uppers,
                'lower_bounds': domain_params.lower_bounds,
                'upper_bounds': domain_params.upper_bounds,
                'slopes': domain_params.slopes if k == 0 else [],
                'cs': domain_params.cs,
                'rhs': domain_params.rhs,
            })

            abs_ret = abstractor._forward_hidden(
                domain_params=k_domain_params,
                decisions=topk_decisions[-1],
                simplify=True
            )
            # improvements over specification
            k_output_lbs = (abs_ret.output_lbs - double_rhs).max(-1).values

            # invalid scores for stable neurons
            # invalid_mask_scores = (topk_scores.values[:, k] <= SMALL).to(torch.get_default_dtype())
            # print(f'{invalid_mask_scores.sum()=}')
            topk_output_lbs[k] = reduce_op((k_output_lbs.flatten()).reshape(2, -1), dim=0).values

        return topk_output_lbs, topk_decisions


    # hidden branching
    # @beartype
    def smart_hidden_branching(self: 'DecisionHeuristic', abstractor: 'abstractor.abstractor.NetworkAbstractor',
                                 domain_params: AbstractResults) -> list[list]:
        batch = len(domain_params.input_lowers)
        topk = min(self.decision_topk, int(sum([i.sum() for (_, i) in domain_params.masks.items()]).item()))
        split_node_names = [_.name for _ in abstractor.net.split_nodes]
        split_node_points = {k: abstractor.net.split_activations[k][0][0].get_split_point() for k in split_node_names}

        # babsr scores
        scores, backup_scores = _compute_babsr_scores(
            abstractor=abstractor,
            lower_bounds=domain_params.lower_bounds,
            upper_bounds=domain_params.upper_bounds,
            lAs=domain_params.lAs,
            batch=batch,
            masks=domain_params.masks,
            reduce_op=self.decision_reduceop,
            number_bounds=domain_params.cs.shape[1]
        )

        # convert an index to its layer and offset
        score_length = np.insert(np.cumsum([len(scores[i][0]) for i in range(len(scores))]), 0, 0)

        # top-k candidates
        topk_scores = torch.topk(torch.cat(scores, dim=1), topk)
        topk_backup_scores = torch.topk(torch.cat(backup_scores, dim=1), topk, largest=False)

        topk_output_lbs, topk_decisions = self.get_topk_scores(
            abstractor=abstractor,
            domain_params=domain_params,
            topk_scores=topk_scores,
            topk_backup_scores=topk_backup_scores,
            score_length=score_length,
            topk=topk,
        )

        # best improvements
        if self.decision_method != 'smart':
            best_output_lbs_indices = np.random.random_integers(low=0, high=len(topk_output_lbs)-1, size=topk_output_lbs.shape[1])
            topk_output_lbs_np = topk_output_lbs.detach().cpu().numpy()
            best_output_lbs = np.array([topk_output_lbs_np[best_output_lbs_indices[ii]][ii] for ii in range(batch * 2)])
        else:
            best = topk_output_lbs.topk(1, 0)
            best_output_lbs = best.values.cpu().numpy()[0]
            best_output_lbs_indices = best.indices.cpu().numpy()[0]

        # align decisions
        all_topk_decisions = [topk_decisions[best_output_lbs_indices[ii]][ii] for ii in range(batch * 2)]
        final_decision = [[] for b in range(batch)]

        for b in range(batch):
            mask_item = {k: domain_params.masks[k][b].clone() for k in split_node_names}
            # valid scores
            if max(best_output_lbs[b], best_output_lbs[b + batch]) > -LARGE:
                n_name, n_id, n_point = all_topk_decisions[b] if best_output_lbs[b] > best_output_lbs[b + batch] else all_topk_decisions[b + batch]
                if n_point is not None: # relu
                    if mask_item[n_name][n_id] != 0: # unstable relu
                        final_decision[b].append([n_name, n_id, n_point])
                        mask_item[n_name][n_id] = 0
                else:
                    assert n_point is None
                    # TODO: general activation
                    raise NotImplementedError
            # invalid scores
            if len(final_decision[b]) == 0:
                # use random decisions
                selected = False
                for layer in np.random.choice(split_node_names, len(split_node_names), replace=False):
                    if (len(mask_item[layer].nonzero(as_tuple=False)) != 0) or (split_node_points[layer] is None):
                        if split_node_points[layer] is not None: # relu
                            final_decision[b].append([layer, mask_item[layer].nonzero(as_tuple=False)[0].item(), split_node_points[layer]])
                            mask_item[final_decision[b][-1][0]][final_decision[b][-1][1]] = 0
                        else:
                            # TODO: general activation
                            raise NotImplementedError
                        selected = True
                        break
                assert selected

        final_decision = sum(final_decision, [])
        return final_decision


    # @beartype
    def naive_input_branching(self: 'DecisionHeuristic', abstractor: 'abstractor.abstractor.NetworkAbstractor',
                        domain_params: AbstractResults) -> torch.Tensor:
        n_inputs = domain_params.input_uppers.flatten(1).shape[1]
        topk = min(self.decision_topk, n_inputs)
        topk_decisions = torch.topk(domain_params.input_uppers.flatten(1) - domain_params.input_lowers.flatten(1), topk, -1).indices
        if topk == 1:
            return topk_decisions

        batch = len(domain_params.input_lowers)
        topk_output_lbs = torch.empty(
            size=(topk, batch),
            device=domain_params.input_lowers.device,
            requires_grad=False,
        )

        for i in range(topk):
            tmp_decision = topk_decisions[:, i:i+1]
            abs_ret = abstractor._forward_input(
                domain_params=domain_params,
                decisions=tmp_decision,
                simplify=True,
            )
            output_lbs_tmp = (abs_ret.output_lbs - torch.cat([domain_params.rhs, domain_params.rhs])).max(-1).values
            # output_lbs_tmp[torch.logical_and(output_lbs_tmp >= -5, output_lbs_tmp > -10)] = LARGE
            topk_output_lbs[i] = torch.max(output_lbs_tmp[:batch], output_lbs_tmp[batch:])

        # print(topk_output_lbs.min(), topk_output_lbs.max())
        # topk_output_lbs[topk_output_lbs >= 0] = -LARGE
        best_output_lbs = torch.topk(topk_output_lbs, 1, 0, largest=True)
        best_indices = best_output_lbs.indices
        # best_values = best_output_lbs.values
        # invalid_indices = torch.where(best_values == -LARGE)

        final_decision = torch.tensor([[topk_decisions[i, best_indices[0, i]]] for i in range(batch)])

        return final_decision


    # @beartype
    def smart_input_branching(self: 'DecisionHeuristic', abstractor: 'abstractor.abstractor.NetworkAbstractor',
                        domain_params: AbstractResults) -> torch.Tensor:
        lAs_factors = domain_params.lAs[abstractor.net.input_name[0]].flatten(2).abs().clamp(min=0.1)
        diff = (domain_params.input_uppers - domain_params.input_lowers).flatten(1).unsqueeze(1)
        objective = (domain_params.output_lbs - domain_params.rhs).unsqueeze(-1)
        score = lAs_factors * diff + objective
        score = score.amax(dim=-2)
        decisions = torch.topk(score, 1, -1).indices
        return decisions


    # @beartype
    def naive_hidden_branching(self: 'DecisionHeuristic', abstractor: 'abstractor.abstractor.NetworkAbstractor', domain_params: AbstractResults, mode: str) -> list[list]:
        batch = len(domain_params.input_lowers)
        split_node_names = [_.name for _ in abstractor.net.split_nodes]
        split_node_points = {k: abstractor.net.split_activations[k][0][0].get_split_point() for k in split_node_names}

        if mode == 'distance':
            scores = {
                k: torch.min(domain_params.upper_bounds[k], -domain_params.lower_bounds[k])
                    for k in split_node_points
            }
        elif mode == 'polarity':
            scores = {
                k: (domain_params.upper_bounds[k] * domain_params.lower_bounds[k]) / (domain_params.lower_bounds[k] - domain_params.upper_bounds[k])
                    for k in split_node_points
            }
        elif mode == 'scale':
            scores = {
                k: torch.min(domain_params.upper_bounds[k], -domain_params.lower_bounds[k]) / torch.abs(domain_params.upper_bounds[k] + domain_params.lower_bounds[k])
                    for k in split_node_points
            }
        elif mode == 'width':
            scores = {
                k: torch.abs(domain_params.upper_bounds[k] - domain_params.lower_bounds[k])
                    for k in split_node_points
            }
        else:
            raise NotImplementedError()

        masks = {
            k: domain_params.masks[k] if (split_node_points[k] is not None) else torch.ones_like(domain_params.masks[k])
                for k in split_node_points
        }

        masked_scores = {
            k: torch.where(masks[k].bool(), scores[k].flatten(1), 0.0)
                for k in split_node_points
        }

        # TODO: not always required to compute
        decision_points = {
            k: (domain_params.upper_bounds[k] + domain_params.lower_bounds[k]) / 2.0
                for k in split_node_points
        }
        assert len(abstractor.net.split_nodes) == len(masked_scores)
        best_scores = [masked_scores[k].topk(1, 1) for k in split_node_points]
        best_scores_all_layers = torch.cat([s.values for s in best_scores], dim=1)
        best_scores_all_layers_indices = torch.cat([s.indices for s in best_scores], dim=1).detach().cpu().numpy()
        best_scores_all = best_scores_all_layers.topk(1, 1)
        assert (best_scores_all.values > 0.0).all()

        layer_ids = best_scores_all.indices[:, 0].detach().cpu().numpy()
        assert len(layer_ids) == batch
        decisions = []
        for b in range(batch):
            l_name = split_node_names[layer_ids[b]]
            n_id = best_scores_all_layers_indices[b, layer_ids[b]]
            if split_node_points[l_name] is not None:
                point = split_node_points[l_name]
            else:
                point = decision_points[l_name][b].flatten()[n_id].item()
            decisions.append([l_name, n_id, point])

        return decisions


    def get_all_branching_rewards(self: 'DecisionHeuristic', abstractor: 'abstractor.abstractor.NetworkAbstractor',
                                 domain_params: AbstractResults, reduce_op=torch.max) -> list[list]:
        # print('get_all_branching_rewards')
        batch = len(domain_params.input_lowers)
        split_node_names = [_.name for _ in abstractor.net.split_nodes]
        split_node_points = {k: abstractor.net.split_activations[k][0][0].get_split_point() for k in split_node_names}

        masks = {
            k: domain_params.masks[k] if (split_node_points[k] is not None) else torch.ones_like(domain_params.masks[k])
                for k in split_node_points
        }

        # print(f'{masks=}')
        n_unstables = sum([v.flatten(1).sum(1) for v in masks.values()])
        assert len(n_unstables) == batch
        topk = int(n_unstables.amax().item())
        topk = topk if self.decision_topk == -1 else max(topk, self.decision_topk)
        # print(f'{topk=} {self.decision_topk=}')
        # print(f'{split_node_names=}')
        # topk = 10

        raw_scores = {
            k: torch.min(domain_params.upper_bounds[k], -domain_params.lower_bounds[k])
                for k in split_node_points
        }

        masked_scores = {
            k: masks[k] * raw_scores[k].flatten(1)
                for k in split_node_points
        }

        # print([s.shape for s in masked_scores.values()])
        scores = [masked_scores[name] for name in split_node_names]

        # convert an index to its layer and offset
        score_length = np.insert(np.cumsum([len(scores[i][0]) for i in range(len(scores))]), 0, 0)

        # top-k candidates
        topk_scores = torch.topk(torch.cat(scores, dim=1), topk)

        topk_output_lbs, topk_decisions = self.get_topk_scores_greedy(
            abstractor=abstractor,
            domain_params=domain_params,
            topk_scores=topk_scores,
            score_length=score_length,
            topk=topk,
            reduce_op=reduce_op,
        )
        # print(f'{batch=}')
        # print(len(topk_output_lbs))
        # print(topk_output_lbs)
        # print(len(topk_decisions))
        # print(topk_decisions)
        return topk_output_lbs, topk_decisions

    # hidden branching
    # @beartype
    def brute_force_hidden_branching(self: 'DecisionHeuristic', abstractor: 'abstractor.abstractor.NetworkAbstractor',
                                 domain_params: AbstractResults) -> list[list]:
        print('brute_force_hidden_branching')
        batch = len(domain_params.input_lowers)
        split_node_names = [_.name for _ in abstractor.net.split_nodes]
        split_node_points = {k: abstractor.net.split_activations[k][0][0].get_split_point() for k in split_node_names}

        masks = {
            k: domain_params.masks[k] if (split_node_points[k] is not None) else torch.ones_like(domain_params.masks[k])
                for k in split_node_points
        }

        # print(f'{masks=}')
        n_unstables = sum([v.flatten(1).sum(1) for v in masks.values()])
        assert len(n_unstables) == batch
        topk = int(n_unstables.amin().item())
        topk = topk if self.decision_topk == -1 else max(topk, self.decision_topk)
        # print(f'{topk=} {self.decision_topk=}')
        # print(f'{split_node_names=}')
        topk = 50

        if 1:
            raw_scores = {
                k: torch.min(domain_params.upper_bounds[k], -domain_params.lower_bounds[k])
                    for k in split_node_points
            }

            masked_scores = {
                k: torch.where(masks[k].bool(), raw_scores[k].flatten(1), 0.0)
                    for k in split_node_points
            }

            # print([s.shape for s in masked_scores.values()])
            scores = [masked_scores[name] for name in split_node_names]
        else:

            scores, backup_scores = _compute_babsr_scores(
                abstractor=abstractor,
                lower_bounds=domain_params.lower_bounds,
                upper_bounds=domain_params.upper_bounds,
                lAs=domain_params.lAs,
                batch=batch,
                masks=domain_params.masks,
                reduce_op=self.decision_reduceop,
                number_bounds=domain_params.cs.shape[1]
            )

        # convert an index to its layer and offset
        score_length = np.insert(np.cumsum([len(scores[i][0]) for i in range(len(scores))]), 0, 0)

        # top-k candidates
        topk_scores = torch.topk(torch.cat(scores, dim=1), topk)
        # topk_backup_scores = torch.topk(torch.cat(backup_scores, dim=1), topk, largest=False)
        # topk_scores_indices = topk_scores.indices.cpu()
        # print(topk_scores_indices.shape)


        topk_output_lbs, topk_decisions = self.get_topk_scores_greedy(
            abstractor=abstractor,
            domain_params=domain_params,
            topk_scores=topk_scores,
            score_length=score_length,
            topk=topk,
        )

        # print(f'{topk_decisions=}')
        # print(f'{topk_output_lbs=}')

        best = topk_output_lbs.topk(1, 0)
        best_output_lbs_indices = best.indices.cpu().numpy()[0]
        all_topk_decisions = [topk_decisions[best_output_lbs_indices[b]][b] for b in range(batch)]
        final_decisions = [[] for b in range(batch)]

        for b in range(batch):
            mask_item = {k: domain_params.masks[k][b].clone() for k in split_node_names}
            n_name, n_id, n_point = all_topk_decisions[b]
            if n_point is not None: # relu
                assert mask_item[n_name][n_id] != 0 # unstable relu
                final_decisions[b].append([n_name, n_id, n_point])
                mask_item[n_name][n_id] = 0
            else:
                # TODO: general activation
                raise NotImplementedError


        final_decisions = sum(final_decisions, [])
        return final_decisions



    def get_branching_scores(self, abstractor, domain_params) -> list[list]:
        batch = len(domain_params.input_lowers)
        split_node_names = [_.name for _ in abstractor.net.split_nodes]
        split_node_points = {k: abstractor.net.split_activations[k][0][0].get_split_point() for k in split_node_names}

        masks = {
            k: domain_params.masks[k] if (split_node_points[k] is not None) else torch.ones_like(domain_params.masks[k])
                for k in split_node_points
        }

        # features
        scores_1, scores_2 = _compute_babsr_scores(
            abstractor=abstractor,
            lower_bounds=domain_params.lower_bounds,
            upper_bounds=domain_params.upper_bounds,
            lAs=domain_params.lAs,
            batch=batch,
            masks=domain_params.masks,
            reduce_op=self.decision_reduceop,
            number_bounds=domain_params.cs.shape[1]
        )
        scores_1 = {split_node_names[i]: scores_1[i] for i in range(len(scores_1))}
        scores_2 = {split_node_names[i]: scores_2[i] for i in range(len(scores_1))}
        scores_3 = {k: torch.min(domain_params.upper_bounds[k], -domain_params.lower_bounds[k]) for k in split_node_names}
        scores_4 = {k: (domain_params.upper_bounds[k] * domain_params.lower_bounds[k]) / (domain_params.lower_bounds[k] - domain_params.upper_bounds[k]) for k in split_node_names}
        scores_5 = {k: torch.min(domain_params.upper_bounds[k], -domain_params.lower_bounds[k]) / torch.abs(domain_params.upper_bounds[k] + domain_params.lower_bounds[k]) for k in split_node_names}
        scores_6 = {k: torch.abs(domain_params.upper_bounds[k] - domain_params.lower_bounds[k]) for k in split_node_names}

        scores_all_dict = {
            k: torch.stack([
                scores_1[k],
                scores_2[k],
                scores_3[k],
                scores_4[k],
                scores_5[k],
                scores_6[k],
            ], dim=-1)
            for k in split_node_names
        }

        scores_all_list = [scores_all_dict[k] for k in split_node_names]
        masks_all_list = [masks[k] for k in split_node_names]
        return scores_all_list, masks_all_list



""" Linear (possibly with weight perturbation) or Dot product layers """
from torch import Tensor
from typing import Tuple, List
from .activation_base import BoundOptimizableActivation
from .base import *
from .bivariate import BoundMul, MulHelper
from .gradient_modules import LinearGrad
from .leaf import BoundParams
from ..patches import Patches, inplace_unfold
from .solver_utils import grb
from .clampmult import multiply_by_A_signs

EPS = 1e-2

class BoundLinear(BoundOptimizableActivation):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        # Gemm:
        # A = A if transA == 0 else A.T
        # B = B if transB == 0 else B.T
        # C = C if C is not None else np.array(0)
        # Y = alpha * np.dot(A, B) + beta * C
        # return Y

        super().__init__(attr, inputs, output_index, options)

        # Defaults in ONNX
        self.transA = 0
        self.transB = 0
        self.alpha_linear = 1.0
        self.beta_linear = 1.0
        if attr is not None:
            self.transA = attr['transA'] if 'transA' in attr else self.transA
            self.transB = attr['transB'] if 'transB' in attr else self.transB
            self.alpha_linear = attr['alpha'] if 'alpha' in attr else self.alpha_linear
            self.beta_linear = attr['beta'] if 'beta' in attr else self.beta_linear

        options = options or {}
        self.opt_matmul = options.get('matmul')
        self.splittable = False

        self.mul_helper = MulHelper()
        self.use_seperate_weights_for_lower_and_upper_bounds = False
        self.share_alphas = options.get('matmul', {}).get('share_alphas', False)

    def _preprocess(self, a, b, c=None):
        """Handle tranpose and linear coefficients."""
        if self.transA and isinstance(a, Tensor):
            a = a.transpose(-2,-1)
        if self.alpha_linear != 1.0:
            a = self.alpha_linear * a
        if not self.transB and isinstance(b, Tensor):
            # our code assumes B is transposed (common case), so we transpose B
            # only when it is not transposed in gemm.
            b = b.transpose(-2, -1)
        if c is not None:
            if self.beta_linear != 1.0:
                c = self.beta_linear * c
        return a, b, c

    def init_opt_parameters(self, start_nodes):
        shared_alpha_dims = []
        if self.share_alphas:
            # TODO Temporarily an adhoc check for alpha sharing.
            count_matmul = len([item for item in self._all_optimizable_activations
                                if isinstance(item, BoundLinear)])
            if count_matmul >= 6:
                shared_alpha_dims = [1, 2, 3]
            elif count_matmul >= 4:
                shared_alpha_dims = [1, 2]

        input_lb = [getattr(xi, 'lower', None) for xi in self.inputs]
        input_ub = [getattr(xi, 'upper', None) for xi in self.inputs]
        input_lb = self._preprocess(*input_lb)
        input_ub = self._preprocess(*input_ub)
        x_l, x_u, y_l, y_u = self._reshape(input_lb[0], input_ub[0], input_lb[1], input_ub[1])
        assert x_l.ndim == y_l.ndim
        shape = [1 if i in shared_alpha_dims
                 else max(x_l.shape[i], y_l.shape[i]) for i in range(x_l.ndim)]
        for start_node in start_nodes:
            ns, size_s = start_node[:2]
            # start_node[3] == False means that this start node is not the final node
            # if not start_node[3]:
            #     # NOTE Experimental code. Please check how it will impact the results.
            #     size_s = 1
            if isinstance(size_s, torch.Size):
                # TODO do not give torch.Size
                size_s = prod(size_s)
            elif isinstance(size_s, (list, tuple)):
                size_s = size_s[0]
            self.alpha[ns] = torch.ones(4, size_s, *shape, device=x_l.device)

    def forward(self, x, w, b=None):
        x, w, b = self._preprocess(x, w, b)
        self.input_shape = self.x_shape = x.shape
        self.y_shape = w.t().shape
        res = x.matmul(w.t())
        if b is not None:
            res += b
        return res

    def onehot_mult(self, weight, bias, C, batch_size):
        """Multiply weight matrix with a diagonal matrix with selected rows."""

        if C is None:
            return None, 0.0

        new_weight = None
        new_bias = 0.0

        if C.index.ndim == 2:
            # Shape is [spec, batch]
            index = C.index.transpose(0, 1)
            coeffs = C.coeffs.transpose(0, 1)
        else:
            index = C.index
            coeffs = C.coeffs

        if C.index.ndim == 1:
            # Every element in the batch shares the same rows.
            if weight is not None:
                new_weight = self.non_deter_index_select(
                    weight, dim=0, index=index
                ).unsqueeze(1).expand(
                    [-1, batch_size] + [-1] * (weight.ndim - 1))
            if bias is not None:
                new_bias = self.non_deter_index_select(
                    bias, dim=0, index=index
                ).unsqueeze(1).expand(-1, batch_size)
        elif C.index.ndim == 2:
            # Every element in the batch has different rows, but the number of
            # rows are the same. This essentially needs a batched index_select function.
            if weight is not None:
                new_weight = batched_index_select(
                    weight.unsqueeze(0), dim=1, index=index)
            if bias is not None:
                new_bias = batched_index_select(
                    bias.unsqueeze(0), dim=1, index=index)
        if C.coeffs is not None:
            if weight is not None:
                new_weight = new_weight * coeffs.unsqueeze(-1)
            if bias is not None:
                new_bias = new_bias * coeffs
        if C.index.ndim == 2:
            # Eventually, the shape of A is [spec, batch, *node] so need a transpose.
            new_weight = new_weight.transpose(0, 1)
            new_bias = new_bias.transpose(0, 1)
        return new_weight, new_bias

    def bound_backward(self, last_lA, last_uA, *x, start_node=None,
                       reduce_bias=True, **kwargs):
        assert len(x) == 2 or len(x) == 3
        if start_node is not None:
            self._start = start_node.name
        has_bias = len(x) == 3
        # x[0]: input node, x[1]: weight, x[2]: bias
        input_lb = [getattr(xi, 'lower', None) for xi in x]
        input_ub = [getattr(xi, 'upper', None) for xi in x]
        # transpose and scale each term if necessary.
        input_lb = self._preprocess(*input_lb)
        input_ub = self._preprocess(*input_ub)
        lA_y = uA_y = lA_bias = uA_bias = None
        lbias = ubias = 0
        batch_size = last_lA.shape[1] if last_lA is not None else last_uA.shape[1]
        weight = input_lb[1]
        bias = input_lb[2] if has_bias else None

        def _bound_oneside(last_A):
            if last_A is None:
                return None, 0
            if isinstance(last_A, torch.Tensor):
                # Matrix mode.
                # Just multiply this layer's weight into bound matrices, and produce biases.
                next_A = last_A.to(weight).matmul(weight)
                sum_bias = (last_A.to(bias).matmul(bias)
                    if has_bias else 0.0)
            elif isinstance(last_A, Patches):
                # Patches mode. After propagating through this layer, it will become a matrix.
                # Reshape the weight matrix as a conv image.
                # Weight was in (linear_output_shape, linear_input_shape)
                # Reshape it to (linear_input_shape, c, h, w)
                reshaped_weight = weight.transpose(0, 1).view(
                    -1, *last_A.input_shape[1:])
                # After unfolding the shape is
                # (linear_input_shape, output_h, output_w, in_c, patch_h, patch_w)
                unfolded_weight = inplace_unfold(
                    reshaped_weight,
                    kernel_size=last_A.patches.shape[-2:],
                    stride=last_A.stride, padding=last_A.padding,
                    inserted_zeros=last_A.inserted_zeros,
                    output_padding=last_A.output_padding)
                if has_bias:
                    # Do the same for the bias.
                    reshaped_bias = bias.view(*last_A.input_shape[1:]).unsqueeze(0)
                    # After unfolding the bias shape is (1, output_h, output_w, in_c, patch_h, patch_w)
                    unfolded_bias = inplace_unfold(
                        reshaped_bias, kernel_size=last_A.patches.shape[-2:],
                        stride=last_A.stride, padding=last_A.padding,
                        inserted_zeros=last_A.inserted_zeros,
                        output_padding=last_A.output_padding)
                if last_A.unstable_idx is not None:
                    # In this case, the last_A shape is (num_unstable, batch, out_c, patch_h, patch_w)
                    # Reshape our weight to (output_h, output_w, 1, in_c, patch_h, patch_w, linear_input_shape), 1 is the inserted batch dim.
                    unfolded_weight_r = unfolded_weight.permute(1, 2, 3, 4, 5, 0).unsqueeze(2)
                    # for sparse patches the shape is (unstable_size, batch, in_c, patch_h, patch_w). Batch size is 1 so no need to select here.
                    # We select in the (output_h, out_w) dimension.
                    selected_weight = unfolded_weight_r[last_A.unstable_idx[1], last_A.unstable_idx[2]]
                    next_A = torch.einsum('sbchw,sbchwi->sbi', last_A.patches, selected_weight)
                    if has_bias:
                        # Reshape our bias to (output_h, output_w, 1, in_c, patch_h, patch_w). We already have the batch dim.
                        unfolded_bias_r = unfolded_bias.permute(1, 2, 0, 3, 4, 5)
                        selected_bias = unfolded_bias_r[last_A.unstable_idx[1], last_A.unstable_idx[2]]
                        sum_bias = torch.einsum('sbchw,sbchw->sb', last_A.patches, selected_bias)
                else:
                    # Reshape our weight to (1, 1, output_h, output_w, in_c, patch_h, patch_w, linear_input_shape), 1 is the spec and batch.
                    selected_weight = unfolded_weight.permute(1, 2, 3, 4, 5, 0).unsqueeze(0).unsqueeze(0)
                    next_A_r = torch.einsum('sbpqchw,sbpqchwi->spqbi', last_A.patches, selected_weight)
                    # We return a matrix with flattened spec dimension (corresponding to out_c * out_h * out_w).
                    next_A = next_A_r.reshape(-1, next_A_r.size(-2), next_A_r.size(-1))
                    if has_bias:
                        # Reshape our bias to (1, 1, output_h, output_w, in_c, patch_h, patch_w)
                        selected_bias = unfolded_bias.unsqueeze(0)
                        sum_bias_r = torch.einsum('sbpqchw,sbpqchw->spqb', last_A.patches, selected_bias)
                        sum_bias = sum_bias_r.reshape(-1, sum_bias_r.size(-1))
            return next_A, sum_bias if has_bias else 0.0

        # Case #1: No weight/bias perturbation, only perturbation on input.
        if not self.is_input_perturbed(1) and (not has_bias or not self.is_input_perturbed(2)):
            # If last_lA and last_uA are indentity matrices.
            # FIXME (12/28): we should check last_lA and last_uA separately.
            # Same applies to the weight perturbed, bias perturbed settings.

            def multiply_with_weight(weight, set_l: bool, set_u: bool):
                lA_x = uA_x = None
                lbias = ubias = 0.
                if isinstance(last_lA, eyeC) and isinstance(last_uA, eyeC):
                    # Use this layer's W as the next bound matrices. Duplicate the batch dimension. Other dimensions are kept 1.
                    # Not perturbed, so we can use either lower or upper.
                    assert last_lA.shape == last_uA.shape
                    shape_others = prod(last_lA.shape[2:-1])
                    A_identity = torch.eye(
                        shape_others, device=weight.device, dtype=weight.dtype
                    ).view(shape_others, 1, 1, shape_others, 1)
                    assert last_lA.shape[0] == weight.size(0) * shape_others
                    w = weight.view(
                        1, weight.size(0), *[1] * (len(last_lA.shape) - 2),
                        weight.size(1))
                    w = w * A_identity

                    # expand the batch_size dim
                    tmp_A_x = w.reshape(
                        last_lA.shape[0], 1, *last_lA.shape[2:-1], weight.size(1)
                    ).expand(last_lA.shape[0], *last_lA.shape[1:-1], weight.size(1))
                    if set_l:
                        lA_x = tmp_A_x
                    if set_u:
                        uA_x = tmp_A_x

                    if has_bias:
                        tmp_bias = bias.unsqueeze(1).repeat(1, batch_size)
                        if set_l:
                            lbias = tmp_bias
                        if set_u:
                            ubias = tmp_bias
                elif isinstance(last_lA, OneHotC) or isinstance(last_uA, OneHotC):
                    # We need to select several rows from the weight matrix
                    # (its shape is output_size * input_size).
                    if set_l:
                        lA_x, lbias = self.onehot_mult(weight, bias, last_lA, batch_size)
                    if last_lA is last_uA and set_l and set_u:
                        uA_x = lA_x
                        ubias = lbias
                    elif set_u:
                        uA_x, ubias = self.onehot_mult(weight, bias, last_uA, batch_size)
                else:
                    if set_l:
                        lA_x, lbias = _bound_oneside(last_lA)
                    if set_u:
                        uA_x, ubias = _bound_oneside(last_uA)
                return lA_x, uA_x, lbias, ubias

            if self.use_seperate_weights_for_lower_and_upper_bounds:
                lA_x, _, lbias, _ = multiply_with_weight(input_lb[1], set_l=True, set_u=False)
                _, uA_x, _, ubias = multiply_with_weight(input_ub[1], set_l=False, set_u=True)
            else:
                lA_x, uA_x, lbias, ubias = multiply_with_weight(weight, set_l=True, set_u=True)

        # Case #2: weight is perturbed. bias may or may not be perturbed.
        elif self.is_input_perturbed(1):
            assert not self.use_seperate_weights_for_lower_and_upper_bounds
            # Obtain relaxations for matrix multiplication.
            [(lA_x, uA_x), (lA_y, uA_y)], lbias, ubias = self.bound_backward_with_weight(
                last_lA, last_uA, input_lb, input_ub, x[0], x[1],
                reduce_bias=reduce_bias, **kwargs)
            if has_bias:
                assert reduce_bias
                if x[2].perturbation is not None:
                    # Bias is also perturbed. Since bias is directly added to the
                    # output, in backward mode it is treated as an input with
                    # last_lA and last_uA as associated bounds matrices.
                    # It's okay if last_lA or last_uA is eyeC, as it will be
                    # handled in the perturbation object.
                    lA_bias = last_lA
                    uA_bias = last_uA
                else:
                    # Bias not perturbed, so directly adding the bias of this
                    # layer to the final bound bias term.
                    if isinstance(last_lA, eyeC) and isinstance(last_uA, eyeC):
                        # Bias will be directly added to output.
                        lbias += input_lb[2].unsqueeze(1).repeat(1, batch_size)
                        ubias += input_lb[2].unsqueeze(1).repeat(1, batch_size)
                    else:
                        if last_lA is not None:
                            lbias += last_lA.matmul(input_lb[2])
                        if last_uA is not None:
                            ubias += last_uA.matmul(input_lb[2])
            # If not has_bias, no need to compute lA_bias and uA_bias
        # Case 3: Only bias is perturbed, weight is not perturbed.
        elif not self.is_input_perturbed(1) and has_bias and self.is_input_perturbed(2):
            assert not self.use_seperate_weights_for_lower_and_upper_bounds
            assert reduce_bias
            if isinstance(last_lA, eyeC) and isinstance(last_uA, eyeC):
                # Use this layer's W as the next bound matrices. Duplicate the
                # batch dimension. Other dimensions are kept 1.
                lA_x = uA_x = input_lb[1].unsqueeze(1).repeat(
                    [1, batch_size] + [1] * (input_lb[1].ndim - 1))
            else:
                lA_x = last_lA.matmul(input_lb[1])
                uA_x = last_uA.matmul(input_lb[1])
            # It's okay if last_lA or last_uA is eyeC, as it will be handled in the perturbation object.
            lA_bias = last_lA
            uA_bias = last_uA
        else:
            assert not self.use_seperate_weights_for_lower_and_upper_bounds

        return [(lA_x, uA_x), (lA_y, uA_y), (lA_bias, uA_bias)], lbias, ubias

    def _reshape(self, x_l, x_u, y_l, y_u):
        x_shape, y_shape = self.input_shape, self.y_shape

        # (x_1, x_2, ..., x_{n-1}, -1, x_n) # FIXME
        x_l = x_l.unsqueeze(-2)
        x_u = x_u.unsqueeze(-2)

        # FIXME merge these two cases
        if len(x_shape) == len(y_shape):
            # (x_1, x_2, ..., -1, y_n, y_{n-1})
            y_l = y_l.unsqueeze(-3)
            y_u = y_u.unsqueeze(-3)
        elif len(y_shape) == 2:
            # (x_1, x_2, ..., -1, y_2, y_1)
            y_l = y_l.reshape(*([1] * (len(x_shape) - 2)), *y_shape).unsqueeze(-3)
            y_u = y_u.reshape(*([1] * (len(x_shape) - 2)), *y_shape).unsqueeze(-3)
        else:
            raise ValueError(f'Unsupported shapes: x_shape {x_shape}, y_shape {y_shape}')

        return x_l, x_u, y_l, y_u

    @staticmethod
    # @torch.jit.script
    def propagate_A_xy(last_A: Tensor, alpha_pos: Tensor, alpha_neg: Tensor,
                       beta_pos: Tensor, beta_neg: Tensor,
                       dim_y: List[int]) -> Tuple[Tensor, Tensor]:
        # last_uA has size (batch, spec, output)
        last_A_pos = last_A.clamp(min=0).unsqueeze(-1)
        last_A_neg = last_A.clamp(max=0).unsqueeze(-1)
        # alpha_u has size (batch, spec, output, input)
        # uA_x has size (batch, spec, input).
        A_x = (alpha_pos.transpose(-1, -2).matmul(last_A_pos) +
                alpha_neg.transpose(-1, -2).matmul(last_A_neg)).squeeze(-1)
        # beta_u has size (batch, spec, output, input)
        # uA_y is for weight matrix, with parameter size (output, input)
        # uA_y has size (batch, spec, output, input). This is an element-wise multiplication.
        # TODO (for zhouxing/qirui): generalize multiply_by_A_signs() to calculate A_x,
        # so last_A_pos and last_A_neg are not needed. This saves memory.
        A_y, _ = multiply_by_A_signs(last_A.unsqueeze(-1), beta_pos, beta_neg, None, None)
        if len(dim_y) != 0:
            A_y = torch.sum(A_y, dim=dim_y)
        return A_x, A_y

    def bound_backward_with_weight(self, last_lA, last_uA, input_lb, input_ub,
                                   x, y, reduce_bias=True, **kwargs):
        # FIXME This is nonlinear. Move to `bivariate.py`.

        # Note: x and y are not tranposed or scaled, and we should avoid using them directly.
        # Use input_lb and input_ub instead.
        (alpha_l, beta_l, gamma_l,
         alpha_u, beta_u, gamma_u) = self.mul_helper.get_relaxation(
             *self._reshape(input_lb[0], input_ub[0], input_lb[1], input_ub[1]),
             self.opt_stage, getattr(self, 'alpha', None),
             getattr(self, '_start', None))
        x_shape = input_lb[0].size()
        if reduce_bias:
            gamma_l = torch.sum(gamma_l, dim=-1)
            gamma_u = torch.sum(gamma_u, dim=-1)

        if len(x.output_shape) != 2 and len(x.output_shape) == len(y.output_shape):
            dim_y = [-3]
        elif len(y.output_shape) == 2:
            dim_y = list(range(2, 2 + len(x_shape) - 2))
        else:
            raise NotImplementedError

        def _bound_oneside(last_A, alpha_pos, beta_pos, gamma_pos, alpha_neg, beta_neg, gamma_neg):
            if last_A is None:
                return None, None, 0
            if isinstance(last_A, eyeC):  # FIXME (12/28): Handle the OneHotC case.
                #FIXME previous implementation is incorrect
                #      expanding eyeC for now
                last_A = (torch.eye(last_A.shape[0], device=last_A.device)
                    .view(last_A.shape[0], 1, *last_A.shape[2:]).expand(last_A.shape))

            A_x, A_y = BoundLinear.propagate_A_xy(
                last_A, alpha_pos, alpha_neg, beta_pos, beta_neg, dim_y)

            if reduce_bias:
                # last_uA has size (batch, spec, output)
                # gamma_u has size (batch, output, 1)
                # ubias has size (batch, spec, 1)
                if self.opt_stage in ['opt', 'reuse']:
                    bias = (torch.einsum('sb...,sb...->sb',
                                        last_A.clamp(min=0), gamma_pos)
                            + torch.einsum('sb...,sb...->sb',
                                        last_A.clamp(max=0), gamma_neg))
                else:
                    bias = (
                        self.get_bias(last_A.clamp(min=0), gamma_pos)
                        + self.get_bias(last_A.clamp(max=0), gamma_neg)
                    )
            else:
                assert self.batch_dim == 0
                assert self.opt_stage not in ['opt', 'reuse']
                assert dim_y == [-3]
                bias = (last_A.unsqueeze(-1).clamp(min=0) * gamma_pos
                        + last_A.unsqueeze(-1).clamp(max=0) * gamma_neg)
                bias_x = bias.sum(dim=-2)
                bias_y = bias.sum(dim=-3)
                bias = (bias_x, bias_y)
            return A_x, A_y, bias

        if self.opt_stage in ['opt', 'reuse']:
            lA_x, lA_y, lbias = _bound_oneside(
                last_lA, alpha_l[0], beta_l[0], gamma_l[0],
                alpha_u[0], beta_u[0], gamma_u[0])
            uA_x, uA_y, ubias = _bound_oneside(
                last_uA, alpha_u[1], beta_u[1], gamma_u[1],
                alpha_l[1], beta_l[1], gamma_l[1])
        else:
            lA_x, lA_y, lbias = _bound_oneside(
                last_lA, alpha_l, beta_l, gamma_l, alpha_u, beta_u, gamma_u)
            uA_x, uA_y, ubias = _bound_oneside(
                last_uA, alpha_u, beta_u, gamma_u, alpha_l, beta_l, gamma_l)

        return [(lA_x, uA_x), (lA_y, uA_y)], lbias, ubias

    @staticmethod
    def _propagate_Linf(x, w):
        h_L, h_U = x
        mid = (h_L + h_U) / 2
        diff = (h_U - h_L) / 2
        w_abs = w.abs()
        if mid.ndim == 2 and w.ndim == 3:
            center = torch.bmm(mid.unsqueeze(1), w.transpose(-1, -2)).squeeze(1)
            deviation = torch.bmm(diff.unsqueeze(1), w_abs.transpose(-1, -2)).squeeze(1)
        else:
            center = mid.matmul(w.transpose(-1, -2))
            deviation = diff.matmul(w_abs.transpose(-1, -2))
        return center, deviation

    def interval_propagate(self, *v, C=None, w=None):
        has_bias = self is not None and len(v) == 3
        if self is not None:
            # This will convert an Interval object to tuple.
            # We need to add perturbation property later.
            v_lb, v_ub = zip(*v)
            v_lb = self._preprocess(*v_lb)
            v_ub = self._preprocess(*v_ub)
            # After preprocess the lower and upper bounds, we make them Intervals again.
            v = [Interval.make_interval(bounds[0], bounds[1], bounds[2])
                 for bounds in zip(v_lb, v_ub, v)]
        if w is None and self is None:
            # Use C as the weight, no bias.
            w, lb, ub = C, torch.tensor(0., device=C.device), torch.tensor(0., device=C.device)
        else:
            if w is None:
                # No specified weight, use this layer's weight.
                if self.is_input_perturbed(1):  # input index 1 is weight.
                    # w is a perturbed tensor. Use IBP with weight perturbation.
                    # C matrix merging not supported.
                    assert C is None
                    res = self.interval_propagate_with_weight(*v)
                    l, u = res
                    if has_bias:
                        return l + v[2][0], u + v[2][1]
                    else:
                        return l, u
                else:
                    # Use weight
                    w = v[1][0]
            if has_bias:
                lb, ub = v[2]
            else:
                lb = ub = 0.0

            if C is not None:
                w = C.matmul(w)
                lb = C.matmul(lb) if not isinstance(lb, float) else lb
                ub = C.matmul(ub) if not isinstance(ub, float) else ub

        # interval_propagate() of the Linear layer may encounter input with different norms.
        norm, eps = Interval.get_perturbation(v[0])[:2]
        if norm == torch.inf:
            interval = BoundLinear._propagate_Linf(v[0], w)
            center, deviation = interval
        elif norm > 0:
            # General Lp norm.
            norm, eps = Interval.get_perturbation(v[0])
            mid = v[0][0]
            dual_norm = np.float64(1.0) / (1 - 1.0 / norm)
            if w.ndim == 3:
                # Extra batch dimension.
                # mid has dimension [batch, input], w has dimension [batch, output, input].
                center = w.matmul(mid.unsqueeze(-1)).squeeze(-1)
            else:
                # mid has dimension [batch, input], w has dimension [output, input].
                center = mid.matmul(w.t())
            deviation = w.norm(dual_norm, dim=-1) * eps
        else:
            # here we calculate the L0 norm IBP bound of Linear layers,
            # using the bound proposed in [Certified Defenses for Adversarial Patches, ICLR 2020]
            norm, eps, ratio = Interval.get_perturbation(v[0])
            mid = v[0][0]
            weight_abs = w.abs()
            if w.ndim == 3:
                # Extra batch dimension.
                # mid has dimension [batch, input], w has dimension [batch, output, input].
                center = w.matmul(mid.unsqueeze(-1)).squeeze(-1)
            else:
                # mid has dimension [batch, input], w has dimension [output, input].
                center = mid.matmul(w.t())
            # L0 norm perturbation
            k = int(eps)
            deviation = torch.sum(torch.topk(weight_abs, k)[0], dim=1) * ratio

        lower, upper = center - deviation + lb, center + deviation + ub

        return (lower, upper)

    def interval_propagate_with_weight(self, *v):
        input_norm, input_eps = Interval.get_perturbation(v[0])
        weight_norm, weight_eps = Interval.get_perturbation(v[1])

        if input_norm == torch.inf and weight_norm == torch.inf:
            # A memory-efficient implementation without expanding all the elementary multiplications
            if self.opt_matmul == 'economic':
                x_l, x_u = v[0][0], v[0][1]
                y_l, y_u = v[1][0].transpose(-1, -2), v[1][1].transpose(-1, -2)

                dx, dy = F.relu(x_u - x_l), F.relu(y_u - y_l)
                base = x_l.matmul(y_l)

                mask_xp, mask_xn = (x_l > 0).to(x_l.dtype), (x_u < 0).to(x_u.dtype)
                mask_xpn = 1 - mask_xp - mask_xn
                mask_yp, mask_yn = (y_l > 0).to(y_l.dtype), (y_u < 0).to(y_u.dtype)
                mask_ypn = 1 - mask_yp - mask_yn

                lower, upper = base.clone(), base.clone()

                lower += dx.matmul(y_l.clamp(max=0)) - (dx * mask_xn).matmul(y_l * mask_ypn)
                upper += dx.matmul(y_l.clamp(min=0)) + (dx * mask_xp).matmul(y_l * mask_ypn)

                lower += x_l.clamp(max=0).matmul(dy) - (x_l * mask_xpn).matmul(dy * mask_yn)
                upper += x_l.clamp(min=0).matmul(dy) + (x_l * mask_xpn).matmul(dy * mask_yp)

                lower += (dx * mask_xn).matmul(dy * mask_yn)
                upper += (dx * (mask_xpn + mask_xp)).matmul(dy * (mask_ypn + mask_yp))
            else:
                # Both input data and weight are Linf perturbed (with upper and lower bounds).
                # We need a x_l, x_u for each row of weight matrix.
                x_l, x_u = v[0][0].unsqueeze(-2), v[0][1].unsqueeze(-2)
                y_l, y_u = v[1][0].unsqueeze(-3), v[1][1].unsqueeze(-3)
                # Reuse the multiplication bounds and sum over results.
                lower, upper = BoundMul.interval_propagate_both_perturbed(*[(x_l, x_u), (y_l, y_u)])
                lower, upper = torch.sum(lower, -1), torch.sum(upper, -1)

            return lower, upper
        elif input_norm == torch.inf and weight_norm == 2:
            # This eps is actually the epsilon per row, as only one row is involved for each output element.
            eps = weight_eps
            # Input data and weight are Linf perturbed (with upper and lower bounds).
            h_L, h_U = v[0]
            # First, handle non-perturbed weight with Linf perturbed data.
            center, deviation = BoundLinear._propagate_Linf(v[0], v[1][0])
            # Compute the maximal L2 norm of data. Size is [batch, 1].
            max_l2 = torch.max(h_L.abs(), h_U.abs()).norm(2, dim=-1).unsqueeze(-1)
            # Add the L2 eps to bounds.
            lb, ub = center - deviation - max_l2 * eps, center + deviation + max_l2 * eps
            return lb, ub
        else:
            raise NotImplementedError(
                "Unsupported perturbation combination: data={}, weight={}".format(input_norm, weight_norm))

    @staticmethod
    @torch.jit.script
    def bound_forward_mul(x_lw: Tensor, x_lb: Tensor, x_uw: Tensor, x_ub: Tensor, w: Tensor):
        w_pos, w_neg = w.clamp(min=0), w.clamp(max=0)
        lw = x_lw.matmul(w_pos) + x_uw.matmul(w_neg)
        uw = x_uw.matmul(w_pos) + x_lw.matmul(w_neg)
        lb = x_lb.matmul(w_pos) + x_ub.matmul(w_neg)
        ub = x_ub.matmul(w_pos) + x_lb.matmul(w_neg)
        return lw, lb, uw, ub

    # w: an optional argument which can be utilized by BoundMatMul
    def bound_dynamic_forward(self, x, w=None, b=None, C=None, max_dim=None, offset=0):
        assert not self.transA and self.alpha_linear == 1.0 and self.transB and self.beta_linear == 1.0
        assert not self.is_input_perturbed(1)
        assert not self.is_input_perturbed(2)

        weight = w.lb
        bias = b.lb if b is not None else None
        if C is not None:
            weight = C.to(weight).matmul(weight).transpose(-1, -2)
            if bias is not None:
                bias = C.to(bias).matmul(bias)
            lb = x.lb.unsqueeze(1)
        else:
            weight = weight.t()
            lb = x.lb

        w_new = x.lw.matmul(weight)
        b_new = lb.matmul(weight)
        if C is not None:
            b_new = b_new.squeeze(1)
        if bias is not None:
            b_new += bias

        return LinearBound(w_new, b_new, w_new, b_new, x_L=x.x_L, x_U=x.x_U, tot_dim=x.tot_dim)

    # w: an optional argument which can be utilized by BoundMatMul
    def bound_forward(self, dim_in, x, w=None, b=None, C=None):
        has_bias = b is not None
        #FIXME _preprocess can only be applied to tensors so far but not linear bounds.
        x, w, b = self._preprocess(x, w, b)

        # Case #1: No weight/bias perturbation, only perturbation on input.
        if not self.is_input_perturbed(1) and (not has_bias or not self.is_input_perturbed(2)):
            if isinstance(w, LinearBound):
                w = w.lower
            if isinstance(b, LinearBound):
                b = b.lower
            if C is not None:
                w = C.to(w).matmul(w).transpose(-1, -2)
                if b is not None:
                    b = C.to(b).matmul(b)
                x_lb, x_ub = x.lb.unsqueeze(1), x.ub.unsqueeze(1)
            else:
                w = w.t()
                x_lb, x_ub = x.lb, x.ub
            lw, lb, uw, ub = BoundLinear.bound_forward_mul(x.lw, x_lb, x.uw, x_ub, w)

            if C is not None:
                lb, ub = lb.squeeze(1), ub.squeeze(1)

            if b is not None:
                lb += b
                ub += b
        # Case #2: weight is perturbed. bias may or may not be perturbed.
        elif self.is_input_perturbed(1):
            if C is not None:
                raise NotImplementedError
            res = self.bound_forward_with_weight(dim_in, x, w)
            if has_bias:
                raise NotImplementedError
            lw, lb, uw, ub = res.lw, res.lb, res.uw, res.ub
        # Case 3: Only bias is perturbed, weight is not perturbed.
        elif not self.is_input_perturbed(1) and has_bias and self.is_input_perturbed(2):
            raise NotImplementedError

        return LinearBound(lw, lb, uw, ub)

    def bound_forward_with_weight(self, dim_in, x, y):
        x_unsqueeze = LinearBound(
            x.lw.unsqueeze(-2), x.lb.unsqueeze(-2),
            x.uw.unsqueeze(-2), x.ub.unsqueeze(-2),
            x.lower.unsqueeze(-2), x.upper.unsqueeze(-2),
        )
        y_unsqueeze = LinearBound(
            y.lw.unsqueeze(-3), y.lb.unsqueeze(-3),
            y.uw.unsqueeze(-3), y.ub.unsqueeze(-3),
            y.lower.unsqueeze(-3), y.upper.unsqueeze(-3),
        )
        res_mul = self.bound_forward_both_perturbed(dim_in, x_unsqueeze, y_unsqueeze)
        return LinearBound(
            res_mul.lw.sum(dim=-1) if res_mul.lw is not None else None,
            res_mul.lb.sum(dim=-1),
            res_mul.uw.sum(dim=-1) if res_mul.uw is not None else None,
            res_mul.ub.sum(dim=-1)
        )

    def build_solver(self, *v, model, C=None, model_type="mip", solver_pkg="gurobi"):
        has_bias = self is not None and len(v) == 3
        # e.g., last layer gurobi vars (1024,)
        gvars_array = np.array(v[0])
        # pre_layer_shape (1024,)
        pre_layer_shape = gvars_array.shape
        # this layer shape (100,)
        # if last layer, this layer shape (9,) instead of (10,)!!!
        this_layer_shape = self.lower.squeeze(0).shape
        out_lbs = self.lower.squeeze(0).detach().cpu().numpy() if self.lower is not None else None
        out_ubs = self.upper.squeeze(0).detach().cpu().numpy() if self.upper is not None else None

        # current layer weight (100, 1024)
        this_layer_weight = v[1]
        if self.transB == 0:
            this_layer_weight = this_layer_weight.transpose(1, 0)
        #### make sure if this is correct for per-label operations
        if C is not None:
            # merge specification C into last layer weights
            # only last layer has C not None
            this_layer_weight = C.squeeze(0).mm(this_layer_weight)
        # if last layer, this layer weight (9,100) instead of (10,100)!!!
        this_layer_weight = this_layer_weight.detach().cpu().numpy()

        this_layer_bias = None
        if has_bias:
            # current layer bias (100,)
            this_layer_bias = v[2]
            if C is not None:
                this_layer_bias = C.squeeze(0).mm(this_layer_bias.unsqueeze(-1)).view(-1)
            # if last layer, this layer bias (9,) instead of (10,)!!!
            this_layer_bias = this_layer_bias.detach().cpu().numpy()

        new_layer_gurobi_vars = []

        for neuron_idx in range(this_layer_shape[0]):
            out_lb = out_lbs[neuron_idx] if out_lbs is not None else -float('inf')
            out_ub = out_ubs[neuron_idx] if out_ubs is not None else float('inf')
            if out_ub - out_lb < EPS:
                """
                    If the inferred lb and ub are too close, it could lead to floating point disagreement
                    between solver's inferred lb and ub constraints and the computed ones from ab-crown.
                    Such disagreement can lead to "infeasible" result from the solver for feasible problem.
                    To avoid so, we relax the box constraints.
                    This should not affect the solver's result correctness,
                    since the tighter lb and ub can be inferred by the solver.
                """
                out_lb, out_ub = (out_lb + out_ub - EPS) / 2., (out_lb + out_ub + EPS) / 2.

            lin_expr = 0
            if has_bias:
                lin_expr = this_layer_bias[neuron_idx].item()
            coeffs = this_layer_weight[neuron_idx, :]

            if solver_pkg == 'gurobi':
                lin_expr += grb.LinExpr(coeffs, v[0])
            else:
                # FIXME (01/12/22): This is slow, must be fixed using addRow() or similar.
                for i in range(len(coeffs)):
                    try:
                        lin_expr += coeffs[i] * v[0][i]
                    except TypeError:
                        lin_expr += coeffs[i] * v[0][i].var

            var = model.addVar(lb=out_lb, ub=out_ub, obj=0,
                                    vtype=grb.GRB.CONTINUOUS,
                                    name=f'lay{self.name}_{neuron_idx}')
            model.addConstr(lin_expr == var, name=f'lay{self.name}_{neuron_idx}_eq')
            new_layer_gurobi_vars.append(var)

        self.solver_vars = new_layer_gurobi_vars
        model.update()

    def build_gradient_node(self, grad_upstream):
        if isinstance(self.inputs[1], BoundParams):
            w = self.inputs[1].param
        else:
            w = self.inputs[1].value
        if not self.transB:
            w = w.t()
        node_grad = LinearGrad(w)
        return node_grad, (grad_upstream,), []

    def update_requires_input_bounds(self):
        self._check_weight_perturbation()


class BoundMatMul(BoundLinear):
    # Reuse most functions from BoundLinear.
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        self.transA = 0
        self.transB = 0
        self.splittable = True
        
    def get_split_point(self):
        return None

    def forward(self, x, y):
        self.x_shape = x.shape
        self.y_shape = y.shape
        self.x = x
        self.y = y
        return x.matmul(y)

    def interval_propagate(self, *v):
        lower, upper = super().interval_propagate(*v)
        return lower, upper

    def bound_backward(self, last_lA, last_uA, *x, start_node=None, **kwargs):
        assert len(x) == 2
        if start_node is not None:
            self._start = start_node.name
        results = super().bound_backward(last_lA, last_uA, *x, **kwargs)
        lA_y = results[0][1][0].transpose(-1, -2) if results[0][1][0] is not None else None
        uA_y = results[0][1][1].transpose(-1, -2) if results[0][1][1] is not None else None
        if isinstance(results[1], tuple):
            lbias = (results[1][0], results[1][1].transpose(-1, -2))
        else:
            lbias = results[1]
        if isinstance(results[2], tuple):
            ubias = (results[2][0], results[2][1].transpose(-1, -2))
        else:
            ubias = results[2]
        return [results[0][0], (lA_y, uA_y), results[0][2]], lbias, ubias

    def bound_forward(self, dim_in, x, y):
        return super().bound_forward(dim_in, x, LinearBound(
            y.lw.transpose(-1, -2) if y.lw is not None else None,
            y.lb.transpose(-1, -2) if y.lb is not None else None,
            y.uw.transpose(-1, -2) if y.uw is not None else None,
            y.ub.transpose(-1, -2) if y.ub is not None else None,
            y.lower.transpose(-1, -2) if y.lower is not None else None,
            y.upper.transpose(-1, -2) if y.upper is not None else None
        ))

    def update_requires_input_bounds(self):
        self.is_linear_op = False
        for inp in self.inputs:
            if not inp.perturbed:
                # If any of the two inputs are constant, we do not need input bounds.
                self.is_linear_op = True
        if self.is_linear_op:
            # One input is constant; no bounds required.
            self.requires_input_bounds = []
            self.splittable = False
        else:
            # Both inputs are perturbed. Need relaxation.
            self.requires_input_bounds = [0, 1]
            self.splittable = True


class BoundNeg(Bound):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        self.ibp_intermediate = True

    def forward(self, x):
        return -x

    def bound_backward(self, last_lA, last_uA, x, **kwargs):
        if type(last_lA) == Tensor or type(last_uA) == Tensor:
            return [(-last_lA if last_lA is not None else None,
                 -last_uA if last_uA is not None else None)], 0, 0
        elif type(last_lA) == Patches or type(last_uA) == Patches:
            if last_lA is not None:
                lA = Patches(-last_lA.patches, last_lA.stride, last_lA.padding,
                             last_lA.shape, unstable_idx=last_lA.unstable_idx,
                             output_shape=last_lA.output_shape)
            else:
                lA = None

            if last_uA is not None:
                uA = Patches(-last_uA.patches, last_uA.stride, last_uA.padding,
                             last_uA.shape, unstable_idx=last_uA.unstable_idx,
                             output_shape=last_uA.output_shape)
            else:
                uA = None
            return [(lA, uA)], 0, 0
        else:
            raise NotImplementedError

    def bound_forward(self, dim_in, x):
        return LinearBound(-x.uw, -x.ub, -x.lw, -x.lb)

    def interval_propagate(self, *v):
        return -v[0][1], -v[0][0]


class BoundCumSum(Bound):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        self.use_default_ibp = True

    def forward(self, x, axis):
        self.axis = axis
        return torch.cumsum(x, axis)

class BoundIdentity(Bound):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        self.use_default_ibp = True

    def forward(self, x):
        return x

    def bound_backward(self, last_lA, last_uA, x, **kwargs):
        return [(last_lA, last_uA)], 0, 0

    def bound_forward(self, dim_in, x):
        return x



import copy
import torch.nn as nn
from torch import Tensor
import torch._C as _C


class BoundedTensor(Tensor):
    @staticmethod
    # We need to override the __new__ method since Tensor is a C class
    def __new__(cls, x, ptb, *args, **kwargs):
        if isinstance(x, Tensor):
            tensor = super().__new__(cls, [], *args, **kwargs)
            tensor.data = x.data
            tensor.requires_grad = x.requires_grad
            return tensor
        else:
            return super().__new__(cls, x, *args, **kwargs)

    def __init__(self, x, ptb):
        self.ptb = ptb

    def __repr__(self):
        if hasattr(self, 'ptb') and self.ptb is not None:
            return '<BoundedTensor: {}, {}>'.format(super().__repr__(), self.ptb.__repr__())
        else:
            return '<BoundedTensor: {}, no ptb>'.format(super().__repr__())

    def clone(self, *args, **kwargs):
        tensor = BoundedTensor(super().clone(*args, **kwargs), copy.deepcopy(self.ptb))
        return tensor

    def _func(self, func, *args, **kwargs):
        temp = func(*args, **kwargs)
        new_obj = BoundedTensor([], self.ptb)
        new_obj.data = temp.data
        new_obj.requires_grad = temp.requires_grad
        return new_obj

    # Copy to other devices with perturbation
    def to(self, *args, **kwargs):
        # FIXME add a general "to" function in perturbation class, not here.
        if hasattr(self.ptb, 'x_L') and isinstance(self.ptb.x_L, Tensor):
            self.ptb.x_L = self.ptb.x_L.to(*args, **kwargs)
        if hasattr(self.ptb, 'x_U') and isinstance(self.ptb.x_U, Tensor):
            self.ptb.x_U = self.ptb.x_U.to(*args, **kwargs)
        if hasattr(self.ptb, 'eps') and isinstance(self.ptb.eps, Tensor):
            self.ptb.eps = self.ptb.eps.to(*args, **kwargs)
        return self._func(super().to, *args, **kwargs)

    @classmethod
    def _convert(cls, ret):
        if cls is Tensor:
            return ret

        if isinstance(ret, Tensor):
            if True:
                # The current implementation does not seem to need non-leaf BoundedTensor
                return ret
            else:
                # Enable this branch if non-leaf BoundedTensor should be kept
                ret = ret.as_subclass(cls)

        if isinstance(ret, tuple):
            ret = tuple(cls._convert(r) for r in ret)

        return ret

    @classmethod
    def __torch_function__(cls, func, types, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}

        if not all(issubclass(cls, t) for t in types):
            return NotImplemented

        with _C.DisableTorchFunction():
            ret = func(*args, **kwargs)
            return cls._convert(ret)


class BoundedParameter(nn.Parameter):
    def __new__(cls, data, ptb, requires_grad=True):
        return BoundedTensor._make_subclass(cls, data, requires_grad)

    def __init__(self, data, ptb, requires_grad=True):
        self.ptb = ptb
        self.requires_grad = requires_grad

    def __deepcopy__(self, memo):
        if id(self) in memo:
            return memo[id(self)]
        else:
            result = type(self)(self.data.clone(), self.ptb, self.requires_grad)
            memo[id(self)] = result
            return result

    def __repr__(self):
        return 'BoundedParameter containing:\n{}\n{}'.format(
            self.data.__repr__(), self.ptb.__repr__())

    def __reduce_ex__(self, proto):
        raise NotImplementedError



from __future__ import annotations
from collections import defaultdict
from beartype import beartype
import typing
import torch
import time

if typing.TYPE_CHECKING:
    from .. import auto_LiRPA

from ..util.misc.tensor_storage import TensorStorage
from ..util.misc.result import AbstractResults
from ..abstractor.utils import _copy_history
from ..heuristic.util import compute_masks
from ..setting import Settings

from ..util.misc.logger import logger

class DomainsList:

    "List of unverified branches"

    # @beartype
    def __init__(self: 'DomainsList',
                 net: auto_LiRPA.BoundedModule | torch._dynamo.OptimizedModule,
                 objective_ids: torch.Tensor,
                 output_lbs: torch.Tensor,
                 input_lowers: torch.Tensor,
                 input_uppers: torch.Tensor,
                 lower_bounds: dict | None,
                 upper_bounds: dict | None,
                 lAs: dict,
                 histories: dict | None,
                 slopes: dict,
                 cs: torch.Tensor,
                 rhs: torch.Tensor,
                 input_split: bool = False,
                 preconditions: dict = {}) -> None:

        self.net = net
        self.final_name = self.net.final_node_name

        self.input_split = input_split
        self.visited = 0
        self.all_conflict_clauses = {int(_): [] for _ in objective_ids}
        self.use_restart = Settings.use_restart and (lower_bounds is not None) and (not input_split)

        # unverified indices
        remain_idx = torch.where((output_lbs.detach().cpu() <= rhs.detach().cpu()).all(1))[0]

        # decisions
        all_histories = [_copy_history(histories) for _ in range(len(cs))] if not input_split else None
        all_betas = [None for i in range(len(cs))] if not input_split else None

        # sat solvers
        self.all_sat_solvers = None
        if self.use_restart and any([len(_) > 0 for _ in preconditions.values()]):
            tic = time.time()
            remain_idx = self.init_sat_solver(
                objective_ids=objective_ids,
                lower_bounds=lower_bounds,
                upper_bounds=upper_bounds,
                histories=all_histories,
                preconditions=preconditions,
                remain_idx=remain_idx,
            )
            logger.info(f'Initialized {len(self.all_sat_solvers)} solvers with {len(preconditions)} learned clauses in {time.time() - tic:.03f} seconds')

        # objective indices
        self.all_objective_ids = TensorStorage(objective_ids[remain_idx].cpu())

        # input bounds
        self.all_input_lowers = TensorStorage(input_lowers[remain_idx].cpu())
        self.all_input_uppers = TensorStorage(input_uppers[remain_idx].cpu())

        # output bounds
        self.all_output_lowers = TensorStorage(output_lbs[remain_idx].cpu())

        # properties
        self.all_cs = TensorStorage(cs[remain_idx].cpu())
        self.all_rhs = TensorStorage(rhs[remain_idx].cpu())

        # alpha
        self.all_slopes = defaultdict(dict)
        for k in slopes:
            self.all_slopes[k] = {}
            for kk, v in slopes[k].items():
                if kk not in self.all_slopes[k]:
                    self.all_slopes[k][kk] = TensorStorage(v[:, :, remain_idx].cpu(), concat_dim=2)
                else:
                    self.all_slopes[k][kk].append(v[:, :, remain_idx].cpu())

        # lAs
        self.all_lAs = {k: TensorStorage(v[remain_idx].cpu()) for k, v in lAs.items()} if len(lAs) else None

        if self.input_split:
            self.all_lower_bounds = self.all_upper_bounds = self.all_histories = self.all_betas = None
        else: # hidden spliting
            # hidden bounds
            self.all_lower_bounds = {k: TensorStorage(v[remain_idx].cpu()) for k, v in lower_bounds.items() if k != self.final_name}
            self.all_upper_bounds = {k: TensorStorage(v[remain_idx].cpu()) for k, v in upper_bounds.items() if k != self.final_name}
            # decisions
            self.all_histories = [all_histories[_] for _ in remain_idx]
            self.all_betas = [all_betas[_] for _ in remain_idx]

        self._check_consistent()


    # @beartype
    @property
    def var_mapping(self: 'DomainsList') -> dict:
        if not hasattr(self, '_var_mapping'):
            self._var_mapping = {}
            count = 1
            for layer in self.net.split_nodes:
                for nid in range(layer.lower.flatten(start_dim=1).shape[-1]):
                    self._var_mapping[layer.name, nid] = count
                    count += 1
        return self._var_mapping


    # @beartype
    @property
    def reversed_var_mapping(self: 'DomainsList') -> dict:
        if not hasattr(self, '_reversed_var_mapping'):
            self._reversed_var_mapping = {v: k for k, v in self.var_mapping.items()}
        return self._reversed_var_mapping


    # @beartype
    def _check_consistent(self: 'DomainsList') -> None:
        # print('Checking domains:', len(self))
        assert len(self.all_input_lowers) == len(self.all_input_uppers) == len(self.all_output_lowers) == len(self), \
            print(len(self.all_input_lowers), len(self.all_input_uppers), len(self.all_output_lowers), len(self))
        assert len(self.all_cs) == len(self.all_rhs) == len(self.all_objective_ids) == len(self), \
            print(len(self.all_cs), len(self.all_rhs), len(self.all_objective_ids))
        assert all([vv.data.shape[2] == len(self) for v in self.all_slopes.values() for vv in v.values()]), \
            print([vv.data.shape[2] for v in self.all_slopes.values() for vv in v.values()], len(self))

        if self.all_lAs is not None:
            assert all([len(_) == len(self) for _ in self.all_lAs.values()])

        if not self.input_split:
            assert len(self.all_betas) == len(self.all_histories) == len(self)
            assert len(self.all_lower_bounds) == len(self.all_upper_bounds)
            assert all([len(_) == len(self) for _ in self.all_lower_bounds.values()])
            assert all([len(_) == len(self) for _ in self.all_upper_bounds.values()])
            if self.all_sat_solvers is not None:
                assert len(self.all_sat_solvers) == len(self), print(f'len(self.all_sat_solvers)={len(self.all_sat_solvers)}, len(self)={len(self)}')


    # @beartype
    def pick_out(self: 'DomainsList', batch: int, device: str = 'cpu') -> AbstractResults:
        assert batch > 0
        batch = min(len(self), batch)
        self.visited += batch

        if torch.cuda.is_available():
            torch.cuda.synchronize()

        # input bounds
        new_input_lowers = self.all_input_lowers.pop(batch).to(device=device, non_blocking=True)
        new_input_uppers = self.all_input_uppers.pop(batch).to(device=device, non_blocking=True)

        # objective indices
        new_objective_ids = self.all_objective_ids.pop(batch).to(device='cpu')

        # output bounds
        new_output_lowers = self.all_output_lowers.pop(batch).to(device=device, non_blocking=True)

        # properties
        new_cs = self.all_cs.pop(batch).to(device=device, non_blocking=True)
        new_rhs = self.all_rhs.pop(batch).to(device=device, non_blocking=True)

        # alpha
        new_slopes = defaultdict(dict)
        for k, v in self.all_slopes.items():
            new_slopes[k] = {kk: vv.pop(batch).to(device=device, non_blocking=True) for (kk, vv) in v.items()}

        # lAs
        new_lAs = {k: lA.pop(batch).to(device=device, non_blocking=True) for (k, lA) in self.all_lAs.items()} if self.all_lAs is not None else None

        if self.input_split:
            # input splitting
            new_lower_bounds = new_upper_bounds = None
            new_masks = new_betas = new_histories = None
            new_sat_solvers = None
        else:
            # hidden spliting
            new_lower_bounds = {k: lb.pop(batch).to(device=device, non_blocking=True) for (k, lb) in self.all_lower_bounds.items()}
            new_upper_bounds = {k: ub.pop(batch).to(device=device, non_blocking=True) for (k, ub) in self.all_upper_bounds.items()}

            # pop batch
            new_betas = self.all_betas[-batch:]
            new_histories = self.all_histories[-batch:]
            # remove batch
            self.all_betas = self.all_betas[:-batch]
            self.all_histories = self.all_histories[:-batch]

            if self.all_sat_solvers is not None:
                new_sat_solvers = self.all_sat_solvers[-batch:]
                # new_sat_solvers = copy.deepcopy(self.all_sat_solvers[-batch:])
                self.all_sat_solvers = self.all_sat_solvers[:-batch]
            else:
                new_sat_solvers = None

            new_masks = compute_masks(
                lower_bounds=new_lower_bounds,
                upper_bounds=new_upper_bounds,
                device=device,
            )

            assert len(new_betas) == len(new_histories) == batch
            assert len(new_input_lowers) == len(new_input_lowers) == batch
            assert len(new_lower_bounds[list(new_lower_bounds.keys())[0]]) == batch
            assert len(new_upper_bounds[list(new_upper_bounds.keys())[0]]) == batch
            assert len(new_lAs[list(new_lAs.keys())[0]]) == batch

        self._check_consistent()

        return AbstractResults(**{
            'objective_ids': new_objective_ids,
            'output_lbs': new_output_lowers,
            'input_lowers': new_input_lowers,
            'input_uppers': new_input_uppers,
            'masks': new_masks,
            'lAs': new_lAs,
            'lower_bounds': new_lower_bounds,
            'upper_bounds': new_upper_bounds,
            'slopes': new_slopes,
            'betas': new_betas,
            'histories': new_histories,
            'cs': new_cs,
            'rhs': new_rhs,
            'sat_solvers': new_sat_solvers,
        })


    # @beartype
    def add(self: 'DomainsList', domain_params: AbstractResults, decisions: list | torch.Tensor) -> None:
        # assert decisions is not None
        batch = len(domain_params.input_lowers)
        assert batch > 0

        # unverified indices
        remaining_index = torch.where((domain_params.output_lbs.detach().cpu() <= domain_params.rhs.detach().cpu()).all(1))[0]

        # hidden splitting
        if not self.input_split:
            # using restart
            if self.all_sat_solvers is not None:
                assert len(domain_params.sat_solvers) == batch
                assert decisions is not None

                extra_conflict_index = []
                for idx_ in remaining_index:
                    # bcp
                    new_sat_solver = self.boolean_propagation(
                        domain_params=domain_params,
                        decisions=decisions,
                        batch_idx=idx_,
                    )
                    if new_sat_solver is None:
                        extra_conflict_index.append(idx_)
                        continue

                    self.all_sat_solvers.append(new_sat_solver)

                if len(extra_conflict_index):
                    logger.debug(f'BCP removes {len(extra_conflict_index)} domains')
                    assert len(extra_conflict_index) == len(list(set(extra_conflict_index)))
                    for eci in extra_conflict_index:
                        remaining_index = remaining_index[remaining_index != eci]

            # decision histories
            self.all_histories.extend([domain_params.histories[i] for i in remaining_index])
            self.all_betas.extend([domain_params.betas[i] for i in remaining_index])

            # conflict clauses
            self.save_conflict_clauses(
                domain_params=domain_params,
                remaining_index=remaining_index,
            )

            # hidden bounds
            [v.append(domain_params.lower_bounds[k][remaining_index]) for k, v in self.all_lower_bounds.items()]
            [v.append(domain_params.upper_bounds[k][remaining_index]) for k, v in self.all_upper_bounds.items()]

        # objective indices
        self.all_objective_ids.append(domain_params.objective_ids[remaining_index])

        # input bounds
        self.all_input_lowers.append(domain_params.input_lowers[remaining_index])
        self.all_input_uppers.append(domain_params.input_uppers[remaining_index])

        # output bounds
        self.all_output_lowers.append(domain_params.output_lbs[remaining_index])

        # properties
        self.all_cs.append(domain_params.cs[remaining_index])
        self.all_rhs.append(domain_params.rhs[remaining_index])

        # alpha
        [vv.append(domain_params.slopes[k][kk][:,:,remaining_index]) for (k, v) in self.all_slopes.items() for (kk, vv) in v.items()]

        # lAs
        [v.append(domain_params.lAs[k][remaining_index]) for k, v in self.all_lAs.items()] if self.all_lAs is not None else None

        # checking
        self._check_consistent()


    # @beartype
    def __len__(self: 'DomainsList') -> int:
        return len(self.all_input_lowers)


    # @beartype
    @property
    def minimum_uppers(self: 'DomainsList') -> float:
        indices = (self.all_output_lowers - self.all_rhs).max(dim=1)[0].argsort()
        if len(indices):
            return (self.all_output_lowers[indices[0]] - self.all_rhs[indices[0]]).max().detach().item()
        return 1e-6

    # @beartype
    @property
    def minimum_lowers(self: 'DomainsList') -> float:
        indices = (self.all_output_lowers - self.all_rhs).min(dim=1)[0].argsort()
        if len(indices):
            return (self.all_output_lowers[indices[0]] - self.all_rhs[indices[0]]).min().detach().item()
        return 1e-6


    # @beartype
    def pick_out_worst_domains(self: 'DomainsList', batch: int, device: str = 'cpu') -> AbstractResults:
        indices = (self.all_output_lowers - self.all_rhs).max(dim=1)[0].argsort()[:batch]

        new_lower_bounds = {k: v[indices].to(device=device, non_blocking=True) for k, v in self.all_lower_bounds.items()}
        new_upper_bounds = {k: v[indices].to(device=device, non_blocking=True) for k, v in self.all_upper_bounds.items()}

        self._check_consistent()

        return AbstractResults(**{
            'lower_bounds': new_lower_bounds,
            'upper_bounds': new_upper_bounds,
        })


    # @beartype
    def update_refined_bounds(self: 'DomainsList', domain_params: typing.Any) -> None:
        # updating
        for key in domain_params.lower_bounds:
            orig_shape = self.all_lower_bounds[key].size()[1:] # skip batch dim

            self.all_lower_bounds[key].copy_(
                torch.where(
                    domain_params.lower_bounds[key].view(orig_shape) > self.all_lower_bounds[key].data,
                    domain_params.lower_bounds[key].view(orig_shape),
                    self.all_lower_bounds[key].data
                )
            )

            self.all_upper_bounds[key].copy_(
                torch.where(
                    domain_params.upper_bounds[key].view(orig_shape) < self.all_upper_bounds[key].data,
                    domain_params.upper_bounds[key].view(orig_shape),
                    self.all_upper_bounds[key].data
                )
            )

        # checking
        self._check_consistent()


    # @beartype
    @torch.no_grad()
    def count_unstable_neurons(self: 'DomainsList') -> torch.Tensor | None:
        if self.all_lower_bounds is None:
            return None

        if not len(self):
            return None

        if not len(self.net.relus):
            return None

        new_masks = compute_masks(
            lower_bounds={k: v.data for k, v in self.all_lower_bounds.items()},
            upper_bounds={k: v.data for k, v in self.all_upper_bounds.items()},
            device='cpu',
        )
        # print(len(self), [_.shape for _ in new_masks.values()])
        n_unstable = sum([_.sum() for _ in new_masks.values()]).int()
        return n_unstable // len(self)


    from .util import init_sat_solver, update_hidden_bounds_histories, boolean_propagation, save_conflict_clauses



import random
from .utils import logger

class BaseScheduler(object):
    def __init__(self, max_eps, opt_str):
        self.parse_opts(opt_str)
        self.prev_loss = self.loss = self.max_eps = self.epoch_length = float("nan")
        self.eps = 0.0
        self.max_eps = max_eps
        self.is_training = True
        self.epoch = 0
        self.batch = 0

    def __repr__(self):
        return '<BaseScheduler: eps {}, max_eps {}>'.format(self.eps, self.max_eps)

    def parse_opts(self, s):
        opts = s.split(',')
        self.params = {}
        for o in opts:
            if o.strip():
                key, val = o.split('=')
                self.params[key] = val

    def get_max_eps(self):
        return self.max_eps

    def get_eps(self):
        return self.eps

    def reached_max_eps(self):
        return abs(self.eps - self.max_eps) < 1e-3

    def step_batch(self, verbose=False):
        if self.is_training:
            self.batch += 1
        return

    def step_epoch(self, verbose=False):
        if self.is_training:
            self.epoch += 1
        return

    def update_loss(self, new_loss):
        self.prev_loss = self.loss
        self.loss = new_loss

    def train(self):
        self.is_training = True

    def eval(self):
        self.is_training = False

    # Set how many batches in an epoch
    def set_epoch_length(self, epoch_length):
        self.epoch_length = epoch_length


class FixedScheduler(BaseScheduler):
    def __init__(self, max_eps, opt_str=""):
        super(FixedScheduler, self).__init__(max_eps, opt_str)
        self.eps = self.max_eps


class LinearScheduler(BaseScheduler):

    def __init__(self, max_eps, opt_str):
        super(LinearScheduler, self).__init__(max_eps, opt_str)
        self.schedule_start = int(self.params['start'])
        self.schedule_length = int(self.params['length'])
        self.epoch_start_eps = self.epoch_end_eps = 0

    def __repr__(self):
        return '<LinearScheduler: start_eps {:.3f}, end_eps {:.3f}>'.format(
            self.epoch_start_eps, self.epoch_end_eps)

    def step_epoch(self, verbose = True):
        self.epoch += 1
        self.batch = 0
        if self.epoch < self.schedule_start:
            self.epoch_start_eps = 0
            self.epoch_end_eps = 0
        else:
            eps_epoch = self.epoch - self.schedule_start
            if self.schedule_length == 0:
                self.epoch_start_eps = self.epoch_end_eps = self.max_eps
            else:
                eps_epoch_step = self.max_eps / self.schedule_length
                self.epoch_start_eps = min(eps_epoch * eps_epoch_step, self.max_eps)
                self.epoch_end_eps = min((eps_epoch + 1) * eps_epoch_step, self.max_eps)
        self.eps = self.epoch_start_eps
        if verbose:
            logger.info("Epoch {:3d} eps start {:7.5f} end {:7.5f}".format(self.epoch, self.epoch_start_eps, self.epoch_end_eps))

    def step_batch(self):
        if self.is_training:
            self.batch += 1
            eps_batch_step = (self.epoch_end_eps - self.epoch_start_eps) / self.epoch_length
            self.eps = self.epoch_start_eps + eps_batch_step * (self.batch - 1)
            if self.batch > self.epoch_length:
                logger.warning('Warning: we expect {} batches in this epoch but this is batch {}'.format(self.epoch_length, self.batch))
                self.eps = self.epoch_end_eps

class RangeScheduler(BaseScheduler):

    def __init__(self, max_eps, opt_str):
        super(RangeScheduler, self).__init__(max_eps, opt_str)
        self.schedule_start = int(self.params['start'])
        self.schedule_length = int(self.params['length'])

    def __repr__(self):
        return '<RangeScheduler: epoch [{}, {}]>'.format(
            self.schedule_start, self.schedule_start + self.schedule_length)

    def step_epoch(self, verbose = True):
        self.epoch += 1
        if self.epoch >= self.schedule_start and self.epoch < self.schedule_start + self.schedule_length:
            self.eps = self.max_eps
        else:
            self.eps = 0

    def step_batch(self):
        pass

class BiLinearScheduler(LinearScheduler):

    def __init__(self, max_eps, opt_str):
        super(BiLinearScheduler, self).__init__(max_eps, opt_str)
        self.schedule_start = int(self.params['start'])
        self.schedule_length = int(self.params['length'])
        self.schedule_length_half = self.schedule_length / 2
        self.epoch_start_eps = self.epoch_end_eps = 0

    def __repr__(self):
        return '<BiLinearScheduler: start_eps {:.5f}, end_eps {:.5f}>'.format(
            self.epoch_start_eps, self.epoch_end_eps)

    def step_epoch(self, verbose = True):
        self.epoch += 1
        self.batch = 0
        if self.epoch < self.schedule_start:
            self.epoch_start_eps = 0
            self.epoch_end_eps = 0
        else:
            eps_epoch = self.epoch - self.schedule_start
            eps_epoch_step = self.max_eps / self.schedule_length_half
            if eps_epoch < self.schedule_length_half:
                self.epoch_start_eps = min(eps_epoch * eps_epoch_step, self.max_eps)
                self.epoch_end_eps = min((eps_epoch + 1) * eps_epoch_step, self.max_eps)
            else:
                self.epoch_start_eps = max(0,
                    self.max_eps - ((eps_epoch - self.schedule_length_half) * eps_epoch_step))
                self.epoch_end_eps = max(0, self.epoch_start_eps - eps_epoch_step)
        self.eps = self.epoch_start_eps
        if verbose:
            logger.info("Epoch {:3d} eps start {:7.5f} end {:7.5f}".format(self.epoch, self.epoch_start_eps, self.epoch_end_eps))


class SmoothedScheduler(BaseScheduler):

    def __init__(self, max_eps, opt_str):
        super(SmoothedScheduler, self).__init__(max_eps, opt_str)
        # Epoch number to start schedule
        self.schedule_start = int(self.params['start'])
        # Epoch length for completing the schedule
        self.schedule_length = int(self.params['length'])
        # Mid point to change exponential to linear schedule
        self.mid_point = float(self.params.get('mid', 0.25))
        # Exponential
        self.beta = float(self.params.get('beta', 4.0))
        assert self.beta >= 2.
        assert self.mid_point >= 0. and self.mid_point <= 1.
        self.batch = 0


    # Set how many batches in an epoch
    def set_epoch_length(self, epoch_length):
        if self.epoch_length != self.epoch_length:
            self.epoch_length = epoch_length
        else:
            if self.epoch_length != epoch_length:
                raise ValueError("epoch_length must stay the same for SmoothedScheduler")

    def step_epoch(self, verbose = True):
        super(SmoothedScheduler, self).step_epoch()
        # FIXME
        if verbose == False:
            for i in range(self.epoch_length):
                self.step_batch()

    # Smooth schedule that slowly morphs into a linear schedule.
    # Code is based on DeepMind's IBP implementation:
    # https://github.com/deepmind/interval-bound-propagation/blob/2c1a56cb0497d6f34514044877a8507c22c1bd85/interval_bound_propagation/src/utils.py#L84
    def step_batch(self, verbose=False):
        if self.is_training:
            self.batch += 1
            init_value = 0.0
            final_value = self.max_eps
            beta = self.beta
            step = self.batch - 1
            # Batch number for schedule start
            init_step = (self.schedule_start - 1) * self.epoch_length
            # Batch number for schedule end
            final_step = (self.schedule_start + self.schedule_length - 1) * self.epoch_length
            # Batch number for switching from exponential to linear schedule
            mid_step = int((final_step - init_step) * self.mid_point) + init_step
            t = (mid_step - init_step) ** (beta - 1.)
            # find coefficient for exponential growth, such that at mid point the gradient is the same as a linear ramp to final value
            alpha = (final_value - init_value) / ((final_step - mid_step) * beta * t + (mid_step - init_step) * t)
            # value at switching point
            mid_value = init_value + alpha * (mid_step - init_step) ** beta
            # return init_value when we have not started
            is_ramp = float(step > init_step)
            # linear schedule after mid step
            is_linear = float(step >= mid_step)
            exp_value = init_value + alpha * float(step - init_step) ** beta
            linear_value = min(mid_value + (final_value - mid_value) * (step - mid_step) / (final_step - mid_step), final_value)
            self.eps = is_ramp * ((1.0 - is_linear) * exp_value + is_linear * linear_value) + (1.0 - is_ramp) * init_value

class AdaptiveScheduler(BaseScheduler):
    def __init__(self, max_eps, opt_str):
        super(AdaptiveScheduler, self).__init__(max_eps, opt_str)
        self.schedule_start = int(self.params['start'])
        self.min_eps_step = float(self.params.get('min_step', 1e-9))
        self.max_eps_step = float(self.params.get('max_step', 1e-4))
        self.eps_increase_thresh = float(self.params.get('increase_thresh', 1.0))
        self.eps_increase_factor = float(self.params.get('increase_factor', 1.5))
        self.eps_decrease_thresh = float(self.params.get('decrease_thresh', 1.5))
        self.eps_decrease_factor = float(self.params.get('decrease_factor', 2.0))
        self.small_loss_thresh = float(self.params.get('small_loss_thresh', 0.05))
        self.epoch = 0
        self.eps_step = self.min_eps_step

    def step_batch(self):
        if self.eps < self.max_eps and self.epoch >= self.schedule_start and self.is_training:
            if self.loss != self.loss or self.prev_loss != self.prev_loss:
                # First 2 steps. Use min eps step
                self.eps += self.min_eps_step
            else:
                # loss decreasing or loss very small. Increase eps step
                if self.loss < self.eps_increase_thresh * self.prev_loss or self.loss < self.small_loss_thresh:
                    self.eps_step = min(self.eps_step * self.eps_increase_factor, self.max_eps_step)
                # loss increasing. Decrease eps step
                elif self.loss > self.eps_decrease_thresh * self.prev_loss:
                    self.eps_step = max(self.eps_step / self.eps_decrease_factor, self.min_eps_step)
                # print("loss {:7.5f} prev_loss {:7.5f} eps_step {:7.5g}".format(self.loss, self.prev_loss, self.eps_step))
                # increase eps according to loss
                self.eps = min(self.eps + self.eps_step, self.max_eps)
            # print("eps step size {:7.5f}, eps {:7.5f}".format(self.eps_step, self.eps))


if __name__ == "__main__":
    s = SmoothedScheduler(0.1, "start=2,length=10,mid=0.3")
    epochs = 20
    batches = 10
    loss = 1.0
    eps = []
    s.set_epoch_length(batches)
    for epoch in range(1,epochs+1):
        s.step_epoch()
        for batch in range(1,batches+1):
            s.step_batch()
            loss = loss * (0.975 + random.random() / 20)
            eps.append(s.get_eps())
            print('epoch {:5d} batch {:5d} eps {:7.5f} loss {:7.5f}'.format(epoch, batch, s.get_eps(), loss))
            # update_loss is only necessary for adaptive eps scheduler
            s.update_loss(loss)
    # plot epsilon values
    import matplotlib
    matplotlib.use('Agg')
    from matplotlib import pyplot as plt
    plt.figure(figsize=(10,8))
    plt.plot(eps)
    plt.xticks(range(0, epochs*batches+batches, batches))
    plt.grid()
    plt.tight_layout()
    plt.savefig('epsilon.pdf')



import warnings

import torch
from torch import nn

from ..utils import is_constant, get_selection, PRINT_DEBUG
from ..operations.base import Operator


class Add(Operator):
    def __init__(self, input_shape=None, input_indices=None, feature_dim=1):
        self.input_shape = input_shape
        self.input_indices = input_indices
        self.feature_dim = feature_dim  # 2 for transformers else 1
        self.out = None

        if input_shape and input_indices:
            self.out = torch.zeros(input_shape)

        super().__init__()

    def forward(self, *input):
        if PRINT_DEBUG:
            print('ADD:', [_.shape for _ in input])

        if self.input_indices:
            out = self.out * 0
            for inp, idx in zip(input, self.input_indices):
                selection = get_selection(idx, self.feature_dim)
                out[selection] += inp
            return out

        # Reorder input so that the matrix is first
        if is_constant(input[0]):
            input = sorted(input, key=lambda x: -len(x.shape))
        # Reorder input so that the broadcasted matrix is last
        elif all(x == 1 for x in input[0].shape):
            input = sorted(input, key=lambda x: -sum(x.shape))

        out = input[0].clone()
        if PRINT_DEBUG:
            print('\t- out:', out.shape)
            print('\t- inputs:', [_.shape for _ in input])
            print('\t- inputs:', input[-1])

        # if out.ndim == 1:
        #     out = out[None]

        for inp in input[1:]:
            # if inp.numel() == 1:
            #     inp = inp.item()
            out += inp
        return out

    def set_input_indices(self, input):
        assert isinstance(input, (list, tuple))

        # If all but one of the inputs are constants do nothing
        # One tensor can easily add together with any number of constants
        if sum(is_constant(inp) for inp in input) >= len(input) - 1:
            return

        input_shape = input[0].shape
        if not all(input_shape == inp.shape for inp in input[1:]):
            warnings.warn("Addition might be corrupted.", RuntimeWarning)
        assert all(
            is_constant(inp) or input_shape[-1] == inp.shape[-1] for inp in input
        )

        # HACK
        while self.feature_dim >= len(input_shape):
            self.feature_dim -= 1
        axis = self.get_axis(input_shape, self.feature_dim)

        input_indices = []
        for inp in input:
            mask = inp != 0
            if len(inp.shape) > 1:
                # Where mask is == 0, the complete input channel can be removed
                s = mask.sum(axis=tuple(axis))
                # If inp is triangular matrix do not remove zero rows.
                # Immediately return.
                seq = torch.arange(len(s))
                if torch.equal(s, seq) or torch.equal(s.flip(0), seq):
                    return
                mask = s != 0
            (non_zeros,) = torch.where(mask)
            input_indices.append(non_zeros)

        # if all elements are non zero, no indices necessary
        if all(len(i) == len(mask) for i in input_indices):
            return

        unique_indices = torch.cat(input_indices).unique()
        input_shape = list(input[0].shape)
        input_shape[self.feature_dim] = len(unique_indices)

        _, input_indices[0] = torch.where(
            input_indices[0][:, None] == unique_indices[None]
        )
        _, input_indices[1] = torch.where(
            input_indices[1][:, None] == unique_indices[None]
        )

        self.input_indices = input_indices
        self.input_shape = tuple(input_shape)
        self.out = nn.Parameter(
            torch.zeros(self.input_shape, device=input[0].device, dtype=input[0].dtype),
            requires_grad=False,
        )

    def __str__(self):
        if self.input_indices:
            return "Add({}, {}, {})".format(
                tuple(self.input_shape),
                len(self.input_indices[0]),
                len(self.input_indices[1]),
            )
        else:
            return "Add(None, None)"



from __future__ import annotations
import warnings
warnings.filterwarnings(action='ignore')
from beartype import beartype
import numpy as np
import logging
import typing
import torch
import time
import copy

from ..onnx2pytorch.convert.model import ConvertModel

from ..heuristic.restart_heuristics import HIDDEN_SPLIT_RESTART_STRATEGIES, INPUT_SPLIT_RESTART_STRATEGIES
from ..heuristic.domains_list import DomainsList

from ..auto_LiRPA.utils import stop_criterion_batch_any

from ..verifier.objective import DnfObjectives
from ..verifier.utils import _prune_domains

from ..abstractor.utils import new_slopes

from ..util.misc.torch_cuda_memory import is_cuda_out_of_memory, gc_cuda
from ..util.misc.result import ReturnStatus
from ..util.misc.logger import logger
from ..util.misc.timer import Timers

from ..setting import Settings


class Verifier:

    "Branch-and-Bound verifier"

    # @beartype
    def __init__(self: 'Verifier', net: 'ConvertModel', input_shape: tuple, batch: int = 1000, device: str = 'cpu') -> None:
        self.net = net # pytorch model
        self.input_shape = input_shape
        self.device = device

        # hyper parameters
        self.input_split = False
        self.batch = max(batch, 1)
        self.orig_batch = max(batch, 1)

        # counter-example
        self.adv = None

        # debug
        self.iteration = 0
        self.last_minimum_lowers = -1e9
        self.tightening_patience = 0

        # stats
        self.all_conflict_clauses = {}
        self.visited = 0


    # @beartype
    def get_objective(self: 'Verifier', dnf_objectives: 'DnfObjectives', max_domain: int):
        # objective = dnf_objectives.pop(1)
        objective = dnf_objectives.pop(max(1, max_domain))
        return objective


    # @beartype
    def compute_stability(self: 'Verifier', dnf_objectives: 'DnfObjectives'):
        print('compute_stability')
        if not (hasattr(self, 'abstractor')):
            self._init_abstractor('backward' if np.prod(self.input_shape) < 100000 else 'forward', dnf_objectives)

        return self.abstractor.compute_stability(dnf_objectives)


    # @beartype
    def verify(self: 'Verifier', dnf_objectives: 'DnfObjectives', preconditions: list = [], timeout: float = 3600.0, force_split: str | None = None) -> str:
        self.start_time = time.time()
        self.status = self._verify(
            dnf_objectives=dnf_objectives,
            preconditions=preconditions,
            timeout=timeout,
            force_split=force_split,
        )
        return self.status


    # @beartype
    def _verify(self: 'Verifier', dnf_objectives: 'DnfObjectives', preconditions: list, timeout: float = 3600.0, force_split: str | None = None) -> str:
        if not len(dnf_objectives):
            return ReturnStatus.UNSAT

        # attack
        Timers.tic('Pre-attack') if Settings.use_timer else None
        is_attacked, self.adv = self._pre_attack(copy.deepcopy(dnf_objectives))
        Timers.toc('Pre-attack') if Settings.use_timer else None
        if is_attacked:
            return ReturnStatus.SAT

        # refine
        Timers.tic('Preprocess') if Settings.use_timer else None
        dnf_objectives, reference_bounds = self._preprocess(dnf_objectives, force_split=force_split)
        Timers.toc('Preprocess') if Settings.use_timer else None
        if not len(dnf_objectives):
            return ReturnStatus.UNSAT

        # mip attack
        is_attacked, self.adv = self._mip_attack(reference_bounds)
        if is_attacked:
            return ReturnStatus.SAT

        status = self._verify_with_restart(
            dnf_objectives=copy.deepcopy(dnf_objectives),
            preconditions=preconditions,
            timeout=timeout,
            reference_bounds=reference_bounds,
            max_domain=self.batch
        )

        if not status:
            status = self._verify_with_restart(
                dnf_objectives=copy.deepcopy(dnf_objectives),
                preconditions=preconditions,
                timeout=timeout,
                reference_bounds=reference_bounds,
                max_domain=1
            )

        return status

    # @beartype
    def _verify_with_restart(self: 'Verifier', dnf_objectives: 'DnfObjectives', preconditions: list,
                             timeout: float = 3600.0, reference_bounds: None | dict = None, max_domain: int = 1) -> str | None:
        # verify
        while len(dnf_objectives):
            Timers.tic('Get objective') if Settings.use_timer else None
            objective = self.get_objective(dnf_objectives, max_domain=max_domain)
            Timers.toc('Get objective') if Settings.use_timer else None

            # restart variables
            nth_restart = 0
            learned_clauses = {int(k): [] for k in objective.ids}
            # TODO: shouldn't add to all objectives
            if len(preconditions): # add to all objective ids
                [learned_clauses[k].extend(preconditions) for k in learned_clauses]

            # verify objective (multiple times if RESTART is returned)
            while True:
                # get strategy + refinement
                Timers.tic('Setup restart') if Settings.use_timer else None
                new_reference_bounds = self._setup_restart(nth_restart, objective)
                Timers.toc('Setup restart') if Settings.use_timer else None

                # adaptive batch size
                while True:
                    logger.info(f'Try batch size {self.batch}')
                    try:
                        # main function
                        status = self._verify_one(
                            objective=objective,
                            preconditions=learned_clauses,
                            reference_bounds=reference_bounds if new_reference_bounds is None else new_reference_bounds,
                            timeout=timeout
                        )
                    except RuntimeError as exception:
                        if is_cuda_out_of_memory(exception):
                            if self.batch == 1:
                                # cannot find a suitable batch size to fit this device
                                logger.debug('[!] OOM with batch_size=1')
                                return ReturnStatus.UNKNOWN
                            self.batch = self.batch // 2
                            dnf_objectives.add(objective)
                            objective = self.get_objective(dnf_objectives, max_domain=max_domain)
                            continue
                        else:
                            # raise NotImplementedError
                            raise
                            logger.debug('[!] RuntimeError exception')
                            return None
                    except SystemExit:
                        exit()
                    except:
                        raise NotImplementedError
                    else:
                        gc_cuda()
                        break

                # stats
                Timers.tic('Save stats') if Settings.use_timer else None
                self._save_stats()
                Timers.toc('Save stats') if Settings.use_timer else None

                # handle returning status
                if status in [ReturnStatus.SAT, ReturnStatus.TIMEOUT, ReturnStatus.UNKNOWN]:
                    return status
                if status == ReturnStatus.UNSAT:
                    break # objective is verified
                if status == ReturnStatus.RESTART:
                    logger.debug('Restarting')
                    # restore original batch size for new restart
                    objective = self._prune_objective(objective)
                    self.batch = self.orig_batch
                    nth_restart += 1
                    # TODO: check general activation
                    if not self.input_split:
                        for k, v in self._get_learned_conflict_clauses().items():
                            learned_clauses[k].extend(v)
                    continue
                raise NotImplementedError()

            logger.info(f'Verified: {len(objective.cs)} \t Remain: {len(dnf_objectives)}')

        return ReturnStatus.UNSAT


    # @beartype
    def _prune_objective(self: 'Verifier', objective: typing.Any) -> typing.Any:
        assert self.domains_list is not None

        all_remaining_ids = torch.unique(self.domains_list.all_objective_ids.data)
        if not len(all_remaining_ids):
            return objective

        # remaining
        indices = torch.tensor([idx for idx, val in enumerate(objective.ids) if val in all_remaining_ids])

        # pruning
        objective.ids = objective.ids[indices]

        objective.lower_bounds = objective.lower_bounds[indices]
        objective.upper_bounds = objective.upper_bounds[indices]

        objective.lower_bounds_f64 = objective.lower_bounds_f64[indices]
        objective.upper_bounds_f64 = objective.upper_bounds_f64[indices]

        objective.cs = objective.cs[indices]
        objective.rhs = objective.rhs[indices]

        objective.cs_f64 = objective.cs_f64[indices]
        objective.rhs_f64 = objective.rhs_f64[indices]

        # assert torch.equal(objective.ids, all_remaining_ids)
        return objective


    # @beartype
    def _initialize(self: 'Verifier', objective, preconditions: dict, reference_bounds: dict | None) -> DomainsList | list:
        # initialization params
        # TODO: fix init_betas found by MIP
        ret = self.abstractor.initialize(objective, reference_bounds=reference_bounds, init_betas=self.refined_betas)

        # check verified
        assert len(ret.output_lbs) == len(objective.cs)
        if stop_criterion_batch_any(objective.rhs.to(self.device))(ret.output_lbs.to(self.device)).all():
            return []

        # full slopes uses too much memory
        slopes = ret.slopes if self.input_split else new_slopes(ret.slopes, self.abstractor.net.final_name)

        # remaining domains
        return DomainsList(
            net=self.abstractor.net,
            objective_ids=ret.objective_ids,
            output_lbs=ret.output_lbs,
            input_lowers=ret.input_lowers,
            input_uppers=ret.input_uppers,
            lower_bounds=ret.lower_bounds,
            upper_bounds=ret.upper_bounds,
            lAs=ret.lAs,
            slopes=slopes, # pruned slopes
            histories=copy.deepcopy(ret.histories),
            cs=ret.cs,
            rhs=ret.rhs,
            input_split=self.input_split,
            preconditions=preconditions,
        )


    # @beartype
    def _verify_one(self: 'Verifier', objective, preconditions: dict, reference_bounds: dict | None, timeout: float) -> str:
        # initialization
        Timers.tic('Initialization') if Settings.use_timer else None
        self.domains_list = self._initialize(objective=objective, preconditions=preconditions, reference_bounds=reference_bounds)
        Timers.toc('Initialization') if Settings.use_timer else None

        # cleaning
        torch.cuda.empty_cache()
        if hasattr(self, 'tightener'):
            self.tightener.reset()

        # main loop
        start_time = time.time()
        start_iteration = self.iteration
        while len(self.domains_list) > 0:
            # search
            Timers.tic('Main loop') if Settings.use_timer else None
            self._parallel_dpll()
            Timers.toc('Main loop') if Settings.use_timer else None

            # check adv founded
            if self.adv is not None:
                if self._check_adv_f64(self.adv, objective):
                    return ReturnStatus.SAT
                logger.debug("[!] Invalid counter-example")
                self.adv = None

            # check timeout
            if self._check_timeout(timeout):
                return ReturnStatus.TIMEOUT

            # check restart
            if self._check_restart(start_time=start_time, start_iteration=start_iteration):
                return ReturnStatus.RESTART

        return ReturnStatus.UNSAT


    # @beartype
    def _check_restart(self: 'Verifier', start_time: float, start_iteration: int) -> bool:
        if not Settings.use_restart:
            return False

        if self.input_split:
            if self.num_restart >= len(INPUT_SPLIT_RESTART_STRATEGIES):
                return False
        else:
            if self.num_restart >= len(HIDDEN_SPLIT_RESTART_STRATEGIES):
                return False

        # restart runtime threshold
        if self.iteration - start_iteration >= 20:
            if time.time() - start_time > Settings.max_restart_runtime:
                logger.debug(f'[Restart] Runtime exceeded {Settings.max_restart_runtime} seconds')
                return True

        # restart domains threshold
        max_branches = Settings.max_input_branches if self.input_split else Settings.max_hidden_branches
        max_visited_branches = Settings.max_input_visited_branches if self.input_split else Settings.max_hidden_visited_branches
        if len(self.domains_list) > max_branches:
            logger.debug(f'[Restart] Number of remaining domains exceeded {max_branches} domains')
            return True

        if self.domains_list.visited > max_visited_branches:
            logger.debug(f'[Restart] Number of visited domains exceeded {max_visited_branches} domains')
            return True

        return False


    # @beartype
    def _parallel_dpll(self: 'Verifier') -> None:
        iter_start = time.time()

        # step 1: MIP attack
        if Settings.use_mip_attack:
            self.mip_attacker.attack_domains(self.domains_list.pick_out_worst_domains(1001, 'cpu'))

        # step 2: stabilizing
        old_domains_length = len(self.domains_list)
        unstable = self.domains_list.count_unstable_neurons()
        if self._check_invoke_tightening(patience_limit=Settings.mip_tightening_patience):
            Timers.tic('Tightening') if Settings.use_timer else None
            self.tightener(
                domain_list=self.domains_list,
                topk=Settings.mip_tightening_topk,
                timeout=Settings.mip_tightening_timeout_per_neuron,
                largest=False, # stabilize near-stable neurons
                solve_both=True, # stabilize both upper and lower bounds
            )
            Timers.toc('Tightening') if Settings.use_timer else None

        # step 3: selection
        Timers.tic('Get domains') if Settings.use_timer else None
        pick_ret = self.domains_list.pick_out(self.batch, self.device)
        Timers.toc('Get domains') if Settings.use_timer else None

        # step 4: PGD attack
        Timers.tic('Loop attack') if Settings.use_timer else None
        self.adv = self._attack(pick_ret, n_interval=Settings.attack_interval, timeout=1.0)
        Timers.toc('Loop attack') if Settings.use_timer else None
        if self.adv is not None:
            return

        # step 5: complete assignments
        self.adv, remain_idx = self._check_full_assignment(pick_ret)
        if (self.adv is not None):
            return

        # pruning
        pruned_ret = _prune_domains(pick_ret, remain_idx) if remain_idx is not None else pick_ret
        if not len(pruned_ret.input_lowers):
            return

        # step 6: branching
        Timers.tic('Decision') if Settings.use_timer else None
        decisions = self.decision(self.abstractor, pruned_ret)
        Timers.toc('Decision') if Settings.use_timer else None

        # step 7: abstraction
        Timers.tic('Abstraction') if Settings.use_timer else None
        abstraction_ret = self.abstractor.forward(decisions, pruned_ret)
        Timers.toc('Abstraction') if Settings.use_timer else None

        # step 8: pruning unverified branches
        Timers.tic('Add domains') if Settings.use_timer else None
        self.domains_list.add(abstraction_ret, decisions)
        Timers.toc('Add domains') if Settings.use_timer else None

        # statistics
        self.iteration += 1
        minimum_lowers = self.domains_list.minimum_lowers
        self._update_tightening_patience(minimum_lowers, old_domains_length)

        # logging
        msg = (
            f'[{"Input" if self.input_split else "Hidden"} splitting]     '
            f'Iteration: {self.iteration:<10} '
            f'Remaining: {len(self.domains_list):<10} '
            f'Visited: {self.domains_list.visited:<10} '
            f'Bound: {minimum_lowers:<15.06f} '
            f'Time elapsed: {time.time() - self.start_time:<10.02f} '
        )
        if logger.level <= logging.DEBUG:
            msg += f'Iteration elapsed: {time.time() - iter_start:<10.02f} '

            if Settings.use_mip_tightening and (not self.input_split):
                msg += f'Tightening patience: {self.tightening_patience}/{Settings.mip_tightening_patience:<10}'

            if (not self.input_split) and (unstable is not None):
                msg += f'Unstable neurons: {unstable:<10}'

        logger.info(msg)


    from .utils import (
        _preprocess, _init_abstractor,
        _check_timeout,
        _setup_restart,
        _pre_attack, _attack, _mip_attack, _check_adv_f64,
        _get_learned_conflict_clauses, _check_full_assignment,
        _check_invoke_tightening, _update_tightening_patience,
        compute_stability, _save_stats, get_stats,
        get_unsat_core,
    )



import time

from termcolor import cprint

class TimerData():
    'Performance timer which can be started with tic() and paused with toc()'

    def __init__(self, name, parent):
        assert parent is None or isinstance(parent, TimerData)

        self.name = name
        self.total_secs = 0
        self.num_calls = 0
        self.last_start_time = None

        self.parent = parent # parent TimerData, None for top-level timers
        self.children = [] # a list of child TimerData

    def get_child(self, name):
        'get a child timer with the given name'

        rv = None

        for child in self.children:
            if child.name == name:
                rv = child
                break

        return rv

    def get_children_recursive(self, name):
        'get all decendants with the given name (returns a list of TimerData)'

        rv = []

        if name == self.name:
            rv.append(self)

        for child in self.children:
            rv += child.get_children_recursive(name)

        return rv

    def full_name(self):
        'get the full name of the timer (including ancestors)'

        if self.parent is None:
            return self.name

        return "{}.{}".format(self.parent.full_name(), self.name)

    def tic(self):
        'start the timer'

        #print "Tic({})".format(self.name)
        if self.last_start_time is not None:
            raise RuntimeError("Timer started twice: {}".format(self.name))

        self.num_calls += 1
        self.last_start_time = time.perf_counter()

    def toc(self):
        'stop the timer'

        #print "Toc({})".format(self.name)

        if self.last_start_time is None:
            raise RuntimeError("Timer stopped without being started: {}".format(self.name))

        self.total_secs += time.perf_counter() - self.last_start_time
        self.last_start_time = None

class Timers():
    '''
    a static class for doing timer messuarements. Use
    Timers.tic(name) and Timers.tic(name) to start and stop timers, use
    print_stats to print time statistics
    '''

    top_level_timer = None
    stack = [] # stack of currently-running timers, parents at the start, children at the end
    enabled = True
    
    def __init__(self):
        raise RuntimeError('Timers is a static class; should not be instantiated')

    @staticmethod
    def reset():
        'reset all timers'

        Timers.top_level_timer = None
        Timers.stack = []

    @staticmethod
    def disable():
        'disables timing measurements'

        Timers.enabled = False

    @staticmethod
    def tic(name):
        'start a timer'

        if Timers.enabled:
            #print(f"> tic({name})")

            if not Timers.stack:
                top = Timers.top_level_timer

                if top is not None and top.name != name:
                    # overwrite old top level timer
                    #print("Overwriting old top-level timer {} with new top-level timer {}".format(top.name, name))
                    top = Timers.top_level_timer = None

                td = top
            else:
                td = Timers.stack[-1].get_child(name)

            # create timer object if it doesn't exist
            if td is None:
                parent = None if not Timers.stack else Timers.stack[-1]
                td = TimerData(name, parent)

                if not Timers.stack:
                    Timers.top_level_timer = td
                else:
                    Timers.stack[-1].children.append(td)

            td.tic()
            Timers.stack.append(td)

    @staticmethod
    def toc(name):
        'stop a timer'

        if Timers.enabled:
            #print(f"< toc({name})")

            if not Timers.stack:
                raise RuntimeError(f"Timer '{name}' was stopped, but no timers were running")

            assert Timers.stack[-1].name == name, "Out of order toc(). Expected to first stop timer {}".format(
                Timers.stack[-1].full_name())

            Timers.stack[-1].toc()
            Timers.stack.pop()
        else:
            assert not Timers.stack, "Timers.enabled was False but Timers.stack non-empty: " + \
                                      f"{[t.name for t in Timers.stack]}"

    @staticmethod
    def print_stats(name=None, short=False):
        'print statistics about performance timers to stdout'

        td = Timers.top_level_timer

        if name is not None and td.name != name:
            if Timers.stack:
                td = Timers.stack[-1]
                
            td = td.get_child(name)
            assert td, f"no direct child with name {name}"

        Timers.print_stats_recursive(td, 0, None, short)

    @staticmethod
    def print_stats_recursive(td, level, total_time, short):
        'recursively print information about a timer'

        assert Timers.enabled, "Timers were disabled"

        low_threshold = 5.0
        high_threshold = 25.0

        if level == 0:
            total_time = td.total_secs

        total_time = 1e-9 if total_time < 1e-9 else total_time # prevent division by zero

        percent_total = 100 * td.total_secs / total_time
        below_threshold = False

        if percent_total < low_threshold:
            below_threshold = True

            if short:
                print_func = lambda _: ""
            else:
                def print_func(text):
                    'below threshold print function'

                    return cprint(text, 'white', attrs=['dark'])
        elif percent_total > high_threshold:
            def print_func(text):
                'above threshold print function'

                return cprint(text, 'red', attrs=['bold'])
        else:
            if short:
                below_threshold = True # short printing has a higher threshold
                print_func = lambda _: ""
            else:
                def print_func(text):
                    'within threshold print function'

                    return cprint(text, None)

        if td.last_start_time is not None:
            raise RuntimeError("Timer was never stopped: {}".format(td.name))

        if td.parent is None or td.parent.total_secs == 0:
            percent = 100
            percent_str = ""
        else:
            percent = 100 * td.total_secs / td.parent.total_secs
            percent_str = " ({:.1f}%)".format(percent)

        print_func("{}{} ({} calls): {:.2f} sec{}".format(" " * level * 2, \
            td.name.capitalize(), td.num_calls, td.total_secs, percent_str))

        if not below_threshold:
            # print children
            total_children_secs = 0

            for child in td.children:
                total_children_secs += child.total_secs

                Timers.print_stats_recursive(child, level + 1, total_time, short)

            if td.children:
                other = td.total_secs - total_children_secs
                other_percent = 100 * other / td.total_secs

                percent_other = 100 * other / total_time

                if percent_other < low_threshold:
                    if short:
                        other_print_func = lambda _: ""
                    else:
                        def other_print_func(text):
                            'below threshold print function'

                            return cprint(text, 'white', attrs=['dark'])
                elif percent_other > high_threshold:
                    def other_print_func(text):
                        'above threshold print function'

                        return cprint(text, 'red', attrs=['bold', 'underline'])
                else:
                    if short:
                        other_print_func = lambda _: ""
                    else:
                        def other_print_func(text):
                            'within threshold print function'

                            return cprint(text, None)
                
                if other_percent > 5.0:
                    percent_str = " ({:.1f}%)".format(other_percent)
                    other_print_func("{}Other ({}): {:.2f} sec{}".format(" " * (level + 1) * 2, td.name.capitalize(), \
                        other, percent_str))


""" Constant operators, including operators that are usually fixed nodes and not perturbed """
from .base import *


class BoundConstant(Bound):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        self.value = attr['value'].to(self.device)
        self.use_default_ibp = True

    def __repr__(self):
        if self.value.numel() == 1:
            return f'BoundConstant(name={self.name}, value={self.value})'
        else:
            return super().__repr__()

    def forward(self):
        return self.value.to(self.device)

    def bound_backward(self, last_lA, last_uA, **kwargs):
        def _bound_oneside(A):
            if A is None:
                return 0.0

            if type(A) == Tensor:
                if A.ndim > 2:
                    A = torch.sum(A, dim=list(range(2, A.ndim)))
            elif type(A) == Patches:
                assert A.padding == 0 or A.padding == (0, 0, 0, 0) or self.value == 0  # FIXME (09/19): adding padding here.
                patches_reshape = torch.sum(A.patches, dim=(-1, -2, -3)) * self.value.to(self.device)
                # Expected shape for bias is (spec, batch, out_h, out_w) or (unstable_size, batch)
                return patches_reshape

            return A * self.value.to(self.device)

        lbias = _bound_oneside(last_lA)
        ubias = _bound_oneside(last_uA)
        return [], lbias, ubias

    def bound_forward(self, dim_in):
        lw = uw = torch.zeros(dim_in, device=self.device)
        lb = ub = self.value
        return LinearBound(lw, lb, uw, ub)

    def build_solver(self, *v, model, C=None, model_type="mip", solver_pkg="gurobi"):
        self.solver_vars = self.value


class BoundPrimConstant(Bound):
    def forward(self):
        return torch.tensor([], device=self.device)

class BoundConstantOfShape(Bound):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        self.value = attr['value'].to(self.device)

    def forward(self, x):
        self.x = x
        self.from_input = True
        return self.value.expand(*list(x))

    def bound_backward(self, last_lA, last_uA, x, **kwargs):
        if last_lA is not None:
            lower_sum_b = last_lA * self.value
            while lower_sum_b.ndim > 2:
                lower_sum_b = torch.sum(lower_sum_b, dim=-1)
        else:
            lower_sum_b = 0

        if last_uA is not None:
            upper_sum_b = last_uA * self.value
            while upper_sum_b.ndim > 2:
                upper_sum_b = torch.sum(upper_sum_b, dim=-1)
        else:
            upper_sum_b = 0

        return [(None, None)], lower_sum_b, upper_sum_b

    def bound_forward(self, dim_in, x):
        assert (len(self.x) >= 1)
        lb = ub = torch.ones(self.output_shape, device=self.device) * self.value
        lw = uw = torch.zeros(self.x[0], dim_in, *self.x[1:], device=self.device)
        return LinearBound(lw, lb, uw, ub)

    def interval_propagate(self, *v):
        self.x = v[0][0]
        value = torch.ones(tuple(v[0][0]), device=self.device) * self.value
        return value, value

    def build_solver(self, *v, model, C=None, model_type="mip", solver_pkg="gurobi"):
        self.solver_vars = self.forward(v)

class BoundRange(Bound):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        self.device = attr['device']

    def forward(self, start, end, step):
        if start.dtype == end.dtype == step.dtype == torch.int64:
            return torch.arange(start, end, step, dtype=torch.int64, device=self.device)
        else:
            return torch.arange(start, end, step, device=self.device)

class BoundATenDiag(Bound):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        self.device = attr['device']

    def forward(self, x, diagonal=0):
        return torch.diag(x, diagonal=diagonal)

    def interval_propagate(self, *v):
        return Interval.make_interval(torch.diag(v[0][0], v[1][0]), torch.diag(v[0][1], v[1][0]), v[0])

class BoundATenDiagonal(Bound):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        self.device = attr['device']

    def forward(self, x, offset=0, dim1=0, dim2=1):
        return torch.diagonal(x, offset=offset, dim1=dim1, dim2=dim2)

    def interval_propagate(self, *v):
        params = (v[1][0], v[2][0], v[3][0])
        return Interval.make_interval(torch.diagonal(v[0][0], *params), torch.diagonal(v[0][1], *params), v[0])
